import inspect

import numpy as np
import pytest

import geomstats.backend as gs

# TODO: remove backend influence
# TODO: simplify most of this


def autograd_backend():
    """Check if autograd is set as backend."""
    return gs.__name__.endswith("autograd")


def np_backend():
    """Check if numpy is set as backend."""
    return gs.__name__.endswith("numpy")


def pytorch_backend():
    """Check if pytorch is set as backend."""
    return gs.__name__.endswith("pytorch")


def autodiff_backend():
    return not np_backend()


if pytorch_backend():
    import torch


autograd_only = pytest.mark.skipif(
    not autograd_backend(), reason="Test for autograd backend only."
)
np_only = pytest.mark.skipif(not np_backend(), reason="Test for numpy backend only.")
torch_only = pytest.mark.skipif(
    not pytorch_backend(), reason="Test for pytorch backends only."
)


np_and_torch_only = pytest.mark.skipif(
    not (np_backend() or pytorch_backend()),
    reason="Test for numpy and pytorch backends only.",
)
np_and_autograd_only = pytest.mark.skipif(
    not (np_backend() or autograd_backend()),
    reason="Test for numpy and autograd backends only.",
)
autograd_and_torch_only = pytest.mark.skipif(
    not (autograd_backend() or pytorch_backend()),
    reason="Test for autograd and torch backends only.",
)


np_autograd_and_torch_only = pytest.mark.skipif(
    not (np_backend() or autograd_backend() or pytorch_backend()),
    reason="Test for numpy, autograd and pytorch backends only.",
)

autodiff_only = autograd_and_torch_only


def pytorch_error_msg(a, b, rtol, atol):
    msg = f"\ntensor 1\n{a}\ntensor 2\n{b}"
    if torch.is_tensor(a) and torch.is_tensor(b):
        if a.dtype == torch.bool and b.dtype == torch.bool:
            diff = torch.logical_xor(a, b)
            msg = msg + f"\ndifference \n{diff}"
        else:
            diff = torch.abs(a - b)
            msg = msg + f"\ndifference \n{diff}\nrtol {rtol}\natol {atol}"
    return msg


def assert_allclose(a, b, rtol=gs.rtol, atol=gs.atol):
    if np_backend() or autograd_backend():
        return np.testing.assert_allclose(a, b, rtol=rtol, atol=atol)

    return assert_true(
        gs.allclose(a, b, rtol=rtol, atol=atol),
        msg=pytorch_error_msg(a, b, rtol, atol),
    )


def assert_true(condition, msg=None):
    assert condition, msg


class TestCase:
    """Class for Geomstats tests."""

    def _test_vectorization(self, vec_data, test_fnc_name=None):
        if test_fnc_name is None:
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            test_fnc_name = calframe[1][3][:-4]

        test_fnc = getattr(self, test_fnc_name)

        for datum in vec_data:
            test_fnc(**datum)

    def assertAllClose(self, a, b, rtol=gs.rtol, atol=gs.atol):
        if (gs.is_array(a) and gs.ndim(a) > 0) or (gs.is_array(b) and gs.ndim(b) > 0):
            self.assertShapeEqual(a, b)
        return assert_allclose(a, b, rtol=rtol, atol=atol)

    def assertAllEqual(self, a, b):
        if np_backend() or autograd_backend():
            np.testing.assert_array_equal(a, b)
        else:
            self.assertTrue(gs.all(gs.equal(a, b)))

    @staticmethod
    def assertTrue(condition, msg=None):
        return assert_true(condition, msg=msg)

    @staticmethod
    def assertFalse(condition, msg=None):
        assert not condition, msg

    @staticmethod
    def assertEqual(a, b, msg=None):
        assert a == b, msg

    @staticmethod
    def assertAllCloseToNp(a, np_a, rtol=gs.rtol, atol=gs.atol):
        are_same_shape = np.all(a.shape == np_a.shape)
        are_same = np.allclose(a, np_a, rtol=rtol, atol=atol)
        assert are_same and are_same_shape

    @staticmethod
    def assertShapeEqual(a, b):
        assert a.shape == b.shape



"""The manifold of full-rank correlation matrices.

Lead authors: Yann Thanwerdas and Olivier Bisson.

References
----------
.. [T2022] Yann Thanwerdas. Riemannian and stratified
    geometries on covariance and correlation matrices. Differential
    Geometry [math.DG]. Université Côte d'Azur, 2022.
"""

import logging

import geomstats.backend as gs
from geomstats.algebra_utils import rowwise_scaling
from geomstats.geometry.base import LevelSet
from geomstats.geometry.diffeo import ComposedDiffeo, Diffeo
from geomstats.geometry.fiber_bundle import DistanceMinimizingAligner, FiberBundle
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.hermitian_matrices import expmh
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.lower_triangular_matrices import StrictlyLowerTriangularMatrices
from geomstats.geometry.manifold import register_quotient
from geomstats.geometry.matrices import (
    Matrices,
    MatricesMetric,
    matrix_matrix_transpose,
    tangent_matrix_matrix_transpose,
)
from geomstats.geometry.open_hemisphere import (
    OpenHemispheresProduct,
    OpenHemisphereToHyperboloidDiffeo,
)
from geomstats.geometry.positive_lower_triangular_matrices import (
    LowerMatrixLog,
    PLTUnitDiagMatrices,
    UnitNormedRowsPLTDiffeo,
)
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.quotient_metric import QuotientMetric
from geomstats.geometry.spd_matrices import (
    CholeskyMap,
    SPDAffineMetric,
    SPDMatrices,
    SymMatrixLog,
    logmh,
)
from geomstats.geometry.symmetric_matrices import (
    HollowMatricesPermutationInvariantMetric,
    NullRowSumsPermutationInvariantMetric,
    NullRowSumsSymmetricMatrices,
    SymmetricHollowMatrices,
)
from geomstats.numerics.optimization import NewtonMethod


def corr_map(point):
    r"""Compute the correlation matrix associated to an SPD matrix.

    .. math::

        \text { Cor }: \Sigma \in \operatorname{Sym}^{+}(n) \longmapsto
        \operatorname{Diag}(\Sigma)^{-1 / 2} \Sigma
        \operatorname{Diag}(\Sigma)^{-1 / 2} \in \operatorname{Cor}^{+}(n)

    Parameters
    ----------
    point : array-like, shape=[..., n, n]
        SPD matrix.

    Returns
    -------
    cor : array_like, shape=[..., n, n]
        Full-rank correlation matrix.
    """
    diagonal = Matrices.diagonal(point) ** (-0.5)
    return FullRankCorrelationMatrices.diag_action(diagonal, point)


def tangent_corr_map(tangent_vec, base_point):
    r"""Compute the differential of the differential of the corr map.

    .. math::

        d_{\Sigma} \operatorname{Cor}(X)=\Delta_{\Sigma}\left[X-\frac{1}{2}
        \left(\Delta_{\Sigma}^2 \operatorname{Diag}(X)
        \Sigma+\Sigma \operatorname{Diag}(X)
        \Delta_{\Sigma}^2\right)\right] \Delta_{\Sigma}

    Parameters
    ----------
    tangent_vec : array-like, shape=[..., n, n]
        Tangent vector.
    base_point : array-like, shape=[..., n, n]
        Base point.

    Returns
    -------
    result : array-like, shape=[..., n, n]
    """
    diagonal_bp = Matrices.diagonal(base_point)
    diagonal_tv = Matrices.diagonal(tangent_vec)

    diagonal = diagonal_tv / diagonal_bp
    aux = base_point * (diagonal[..., None, :] + diagonal[..., :, None])
    mat = tangent_vec - 0.5 * aux

    scaled_diag_vec = diagonal_bp ** (-0.5)

    return FullRankCorrelationMatrices.diag_action(scaled_diag_vec, mat)


class FullRankCorrelationMatrices(LevelSet):
    """Class for the manifold of full-rank correlation matrices.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(dim=int(n * (n - 1) / 2), equip=equip)

    def _define_embedding_space(self):
        return SPDMatrices(n=self.n)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return FullRankCorrelationAffineQuotientMetric

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n]
        """
        return Matrices.diagonal(point) - gs.ones(self.n)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : Ignored.

        Returns
        -------
        submersed_vector : array-like, shape=[..., n]
        """
        submersed_vector = Matrices.diagonal(vector)
        if point is not None and point.ndim > vector.ndim:
            return gs.broadcast_to(submersed_vector, point.shape[:-1])

        return submersed_vector

    @staticmethod
    def diag_action(diagonal_vec, point):
        r"""Apply a diagonal matrix on an SPD matrices by congruence.

        The action of :math:`D` on :math:`\Sigma` is given by :math:`D
        \Sigma D`. The diagonal matrix must be passed as a vector representing
        its diagonal.

        Parameters
        ----------
        diagonal_vec : array-like, shape=[..., n]
            Vector coefficient of the diagonal matrix.
        point : array-like, shape=[..., n, n]
            Symmetric Positive definite matrix.

        Returns
        -------
        mat : array-like, shape=[..., n, n]
            Symmetric matrix obtained by the action of `diagonal_vec` on
            `point`.
        """
        return point * gs.outer(diagonal_vec, diagonal_vec)

    @classmethod
    def from_covariance(cls, point):
        r"""Compute the correlation matrix associated to an SPD matrix.

        The correlation matrix associated to an SPD matrix (the covariance)
        :math:`\Sigma` is given by :math:`D \Sigma D` where :math:`D` is
        the inverse square-root of the diagonal of :math:`\Sigma`.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Symmetric Positive definite matrix.

        Returns
        -------
        corr : array-like, shape=[..., n, n]
            Correlation matrix obtained by dividing all elements by the
            diagonal entries.
        """
        diag_vec = Matrices.diagonal(point) ** (-0.5)
        return cls.diag_action(diag_vec, point)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample full-rank correlation matrices by projecting random SPD mats.

        Parameters
        ----------
        n_samples : int
            Number of samples.
        bound : float
            Bound of the interval in which to sample.
            Optional, default: 1.

        Returns
        -------
        cor : array-like, shape=[n_samples, n, n]
            Sample of full-rank correlation matrices.
        """
        spd = self.embedding_space.random_point(n_samples, bound=bound)
        return self.from_covariance(spd)

    def projection(self, point):
        """Project a matrix to the space of correlation matrices.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to project.
        """
        spd = self.embedding_space.projection(point)
        return self.from_covariance(spd)

    def to_tangent(self, vector, base_point):
        """Project a matrix to the tangent space at a base point.

        The tangent space to the space of correlation matrices is the space of
        symmetric matrices with null diagonal.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
            Matrix to project
        base_point : array-like, shape=[..., n, n]
            Correlation matrix.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Symmetric matrix with 0 diagonal.
        """
        sym = self.embedding_space.to_tangent(vector, base_point)
        mask_diag = gs.ones_like(vector) - gs.eye(self.n)
        return sym * mask_diag


class CorrelationMatricesBundle(FiberBundle):
    """Fiber bundle to construct the quotient metric on correlation matrices.

    Correlation matrices are obtained as the quotient of the space of SPD
    matrices by the action by congruence of diagonal matrices.

    References
    ----------
    .. [TP21] Thanwerdas, Yann, and Xavier Pennec. “Geodesics and Curvature of
        the Quotient-Affine Metrics on Full-Rank CorrelationMatrices.”
        In Proceedings of Geometric Science of Information.
        Paris, France, 2021.
        https://hal.archives-ouvertes.fr/hal-03157992.
    """

    def __init__(self, total_space):
        aligner = (
            DistanceMinimizingAligner(total_space, group_elem_shape=(total_space.n,))
            if gs.has_autodiff()
            else None
        )

        super().__init__(total_space=total_space, aligner=aligner)

    @staticmethod
    def riemannian_submersion(point):
        """Compute the correlation matrix associated to an SPD matrix.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            SPD matrix.

        Returns
        -------
        cor : array_like, shape=[..., n, n]
            Full rank correlation matrix.
        """
        return corr_map(point)

    def tangent_riemannian_submersion(self, tangent_vec, base_point):
        """Compute the differential of the submersion.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        result : array-like, shape=[..., n, n]
        """
        return tangent_corr_map(tangent_vec, base_point)

    def vertical_projection(self, tangent_vec, base_point):
        """Compute the vertical projection wrt the affine-invariant metric.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        ver : array-like, shape=[..., n, n]
            Vertical projection.
        """
        n = self._total_space.n
        inverse_base_point = GeneralLinear.inverse(base_point)
        operator = gs.eye(n) + base_point * inverse_base_point
        inverse_operator = GeneralLinear.inverse(operator)
        vector = gs.einsum("...ij,...ji->...i", inverse_base_point, tangent_vec)
        diagonal = gs.einsum("...ij,...j->...i", inverse_operator, vector)
        return base_point * (diagonal[..., None, :] + diagonal[..., :, None])

    def horizontal_lift(self, tangent_vec, base_point=None, fiber_point=None):
        """Compute the horizontal lift wrt the affine-invariant metric.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector of the manifold of full-rank correlation matrices.
        fiber_point : array-like, shape=[..., n, n]
            SPD matrix in the fiber above point.
        base_point : array-like, shape=[..., n, n]
            Full-rank correlation matrix.

        Returns
        -------
        hor_lift : array-like, shape=[..., n, n]
            Horizontal lift of tangent_vec from point to base_point.
        """
        if base_point is not None:
            return self.horizontal_projection(tangent_vec, base_point)
        diagonal_point = Matrices.diagonal(fiber_point) ** 0.5
        lift = self._total_space.group_action(diagonal_point, tangent_vec)
        return self.horizontal_projection(lift, base_point=fiber_point)


class FullRankCorrelationAffineQuotientMetric(QuotientMetric):
    """Class for the quotient of the affine-invariant metric.

    The affine-invariant metric on SPD matrices is invariant under the
    action of diagonal matrices, thus it induces a quotient metric on the
    manifold of full-rank correlation matrices.
    """

    def __init__(self, space, total_space=None):
        if total_space is None:
            total_space = SPDMatrices(space.n, equip=False)
            total_space.equip_with_metric(SPDAffineMetric)

        if not hasattr(total_space, "group_action"):
            total_space.equip_with_group_action(FullRankCorrelationMatrices.diag_action)

        if not hasattr(total_space, "quotient"):
            total_space.equip_with_quotient()

        super().__init__(
            space=space,
            total_space=total_space,
        )


class PolyHyperbolicCholeskyMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism between full-rank correlation matrices and
    the space of lower triangular matrices with positive diagonal
    and unit normed rows.

    Since this image space is also diffeomorphic to another space, the
    product space of successively increasing factor-dimension open hemispheres,
    we take advantage of `ComposedDiffeo` to avoid explicitly representing
    the intermediate space.

    For more details, check section 7.4.1 [T2022]_.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space):
        n = space.n
        diffeos = [CholeskyMap(), UnitNormedRowsPLTDiffeo(n)]

        if n == 2:
            diffeos.append(OpenHemisphereToHyperboloidDiffeo())
            image_space = Hyperboloid(dim=1)
        else:
            image_space = OpenHemispheresProduct(n=n)

        diffeo = ComposedDiffeo(diffeos)

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


def off_map(matrix):
    """Subtract diagonal to a matrix."""
    return matrix - Matrices.to_diagonal(matrix)


class EuclideanCholeskyDiffeo(Diffeo):
    r"""Euclidean-Cholesky diffeomorphism from Cor+ to LT^1.

    A diffeomorphism between full-rank correlation matrices
    :math:`\mathrm{Cor}(n)` and lower triangular matrices with
    unit diagonal :math:`\mathrm{LT}^1(n)`.

    Check out chapter 7 of [T2022]_ for more details.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __call__(self, base_point):
        r"""Diffeomorphism at base point.

        .. math::

           \Theta(C) = \operatorname{Diag}(\operatorname{Chol}(C))^{-1}
           \operatorname{Chol}(C)

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., n, n]
            Image point.
        """
        chol = gs.linalg.cholesky(base_point)
        diag_inv = 1 / Matrices.diagonal(chol)
        return rowwise_scaling(diag_inv, chol)

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        .. math::

            \begin{aligned}
            \Phi(\Gamma) &= \operatorname{Diag}\left(\phi(\Gamma)\right)^{-1 / 2} 
            \phi(\Gamma) \operatorname{Diag}\left(\phi(\Gamma)\right)^{-1 / 2} \\
            &= \operatorname{Cor} \circ \phi(\Gamma)
            \end{aligned}

        Parameters
        ----------
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., n, n]
            Base point.
        """
        return corr_map(matrix_matrix_transpose(image_point))

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        .. math::

            d_C \Theta (X)=\Theta(C) \operatorname{Low}_s\left(L^{-1} X L^{-\top}\right)
            -\frac{1}{2} \operatorname{Diag}\left(L^{-1} X L^{-\top}\right) \Theta(C)

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image of the base point.
        """
        if base_point is None:
            base_point = self.inverse(image_point)

        if image_point is None:
            image_point = self(base_point)

        chol = gs.linalg.cholesky(base_point)
        chol_inv = gs.linalg.inv(chol)
        aux = Matrices.mul(chol_inv, tangent_vec, Matrices.transpose(chol_inv))
        term_1 = Matrices.mul(
            image_point,
            Matrices.to_lower_triangular_diagonal_scaled(aux),
        )
        term_2 = 1 / 2 * rowwise_scaling(Matrices.diagonal(aux), image_point)

        return term_1 - term_2

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        .. math::

            d_{\Gamma} \Phi (Y)= d_{\phi(\Gamma)} \operatorname{Cor}
            \circ d_{\Gamma} \phi(Y)

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., n, n]
            Image point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        if image_point is None:
            image_point = self(base_point)

        image_point_ = matrix_matrix_transpose(image_point)
        image_tangent_vec_ = tangent_matrix_matrix_transpose(
            image_tangent_vec, image_point
        )
        return tangent_corr_map(image_tangent_vec_, image_point_)


class EuclideanCholeskyMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism is between full-rank correlation matrices and
    lower triangular matrices with unit diagonal endowed with a
    Euclidean metric.

    For more details, check section 7.4.2 of [T2022]_.

    Parameters
    ----------
    space : FullRankCorrelationMatrices
        Manifold.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space):
        diffeo = EuclideanCholeskyDiffeo()
        image_space = PLTUnitDiagMatrices(n=space.n, equip=False).equip_with_metric(
            MatricesMetric
        )

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


class LogEuclideanCholeskyDiffeo(ComposedDiffeo):
    r"""log-Euclidean-Cholesky diffeomorphism from Cor+ to LT^0.

    A diffeomorphism between full-rank correlation matrices
    :math:`\mathrm{Cor}(n)` and strictly lower triangular matrices
    :math:`\mathrm{LT}^0(n)`.

    Check out chapter 7 of [T2022]_ for more details.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self):
        diffeos = [EuclideanCholeskyDiffeo(), LowerMatrixLog()]
        super().__init__(diffeos=diffeos)


class LogEuclideanCholeskyMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism is between full-rank correlation matrices and
    strictly lower triangular matrices endowed with a
    Euclidean metric.

    For more details, check section 7.4.3 of [T2022]_.

    Parameters
    ----------
    space : FullRankCorrelationMatrices
        Manifold.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space):
        diffeo = LogEuclideanCholeskyDiffeo()
        image_space = StrictlyLowerTriangularMatrices(
            n=space.n, equip=False
        ).equip_with_metric(MatricesMetric)

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


class UniqueDiagonalMatrixAlgorithm:
    r"""Find unique diagonal matrix corresponding to a full-rank correlation matrix.

    That is, for all symmetric matrix :math:`S`,
    there exists a unique diagonal matrix :math:`D` such that
    :math:`expm(D+S)` is a full-rank correlation matrix.

    Converges in logarithmic time to the solution of the equation, no closed form.

    Check out Theorem 8.10 of [T2022]_ for more details.

    Parameters
    ----------
    atol : float
        Tolerance to check algorithm convergence.
    max_iter : int
        Maximum iterations.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    .. [AH2020] Ilya Archakov, and Peter Reinhard Hansen.
        “A New Parametrization of Correlation Matrices.” arXiv, December 3, 2020.
        https://doi.org/10.48550/arXiv.2012.02395.
    """

    def __init__(self, atol=gs.atol, max_iter=100):
        self.atol = atol
        self.max_iter = max_iter

    def _check_convergence(self, new_matrix, matrix):
        mat_diff = new_matrix - matrix
        return gs.linalg.norm(mat_diff, axis=(-2, -1)) < self.atol

    def _call_single(self, sym_mat):
        r"""Find unique diagonal matrix corresponding to a full-rank correlation matrix.

        Parameters
        ----------
        sym_mat : array-like, shape=[n, n]

        Returns
        -------
        diag_mat : array-like, shape=[n, n]
        """
        diag_mat = gs.zeros_like(sym_mat)

        for _ in range(self.max_iter):
            approx_cor_mat = expmh(diag_mat + sym_mat)
            new_diag_mat = diag_mat - logmh(Matrices.to_diagonal(approx_cor_mat))

            if self._check_convergence(new_diag_mat, diag_mat):
                return diag_mat

            diag_mat = new_diag_mat
        else:
            logging.warning(
                "Maximum number of iterations %d reached. The mean may be inaccurate",
                self.max_iter,
            )

        return diag_mat

    def __call__(self, sym_mat):
        r"""Find unique diagonal matrix corresponding to a full-rank correlation matrix.

        Parameters
        ----------
        sym_mat : array-like, shape=[..., n, n]

        Returns
        -------
        diag_mat : array-like, shape=[..., n, n]
        """
        if sym_mat.ndim == 2:
            return self._call_single(sym_mat)

        batch_shape = sym_mat.shape[:-2]
        if len(batch_shape) == 1:
            return gs.stack([self._call_single(sym_mat_) for sym_mat_ in sym_mat])

        mat_shape = sym_mat.shape[-2:]
        flat_sym_mat = gs.reshape(sym_mat, (-1,) + mat_shape)
        out = gs.stack([self._call_single(sym_mat_) for sym_mat_ in flat_sym_mat])
        return gs.reshape(out, batch_shape + mat_shape)


class OffLogDiffeo(Diffeo):
    r"""Off-log diffeomorphism from Cor+ to Hol.

    A diffeomorphism between full-rank correlation matrices Cor+ and
    symmetric hollow matrices Hol:

    .. math::
        \operatorname{Log} = \operatorname{Off} \circ \log :
        \operatorname{Cor}^{+}(n) \longrightarrow \operatorname{Hol}(n)

    Check out chapter 8 of [T2022]_ for more details.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self):
        super().__init__()
        self.unique_diag_mat = UniqueDiagonalMatrixAlgorithm()

    def __call__(self, base_point):
        """Diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., n, n]
            Image point.
        """
        return off_map(matrix=logmh(mat=base_point))

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., n, n]
            Base point.
        """
        return expmh(self.unique_diag_mat(image_point) + image_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image of the base point.
        """
        if base_point is None:
            base_point = self.inverse(image_point)

        return off_map(
            SymMatrixLog.tangent(tangent_vec=tangent_vec, base_point=base_point)
        )

    def _divided_difference_exp(self, eigvals):
        r"""First divided difference function of the exponential, :math:`exp^(1)`.

        If :math:` x \neq y`,

        .. math::

            exp^(1) = (exp(x)-exp(y))/(x-y)

        else:

        .. math::

            exp'(x)=exp(x)

        Parameters
        ----------
        eigvals : array-like, shape=[..., n]
            Typically eigenvalues of the matrix.

        Returns
        -------
        divided_diffs : array-like, shape=[..., n, n]
            First divided difference function of the exponential.

        """
        eigvals_ = gs.expand_dims(eigvals, axis=-2)
        eigvals_t = gs.expand_dims(eigvals, axis=-1)

        eigvals_diff = eigvals_ - eigvals_t

        mask = gs.logical_and(-gs.atol < eigvals_diff, eigvals_diff < gs.atol)

        exp_eigvals = gs.exp(eigvals)

        exp_eigvals_ = gs.expand_dims(exp_eigvals, axis=-2)
        exp_eigvals_t = gs.expand_dims(exp_eigvals, axis=-1)
        default_vals = exp_eigvals_ - gs.zeros((eigvals.shape[-1], 1))

        return gs.where(
            mask,
            default_vals,
            gs.divide(exp_eigvals_ - exp_eigvals_t, eigvals_diff, ignore_div_zero=True),
        )

    def _build_tangent_diag_aux_mat(self, image_point=None, base_point=None):
        r"""Build auxiliar matrix for tangent diagonal map computation.

        The :math:`H_0` matrix is a SPD matrix where each coefficient is

        .. math::

            (H_0)_il = \sum_{j,k} P_ij*P_ik*P_lj*P_lk*exp^(1)(d_j, d_k)

        where :math:`PDP^t = D+S`, with :math:`D` being a diagonal matrix obtained
        using `UniqueDiagonalMatrixAlgorithm` and :math:`S` is a hollow matrix.

        It is used to compute the pushforward of the
        `UniqueDiagonalMatrixAlgorithm` application.

        Parameters
        ----------
        image_point : array-like, shape=[..., n, n]
            Image of base point by the diffeomorphism.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        h0_mat : array-like, shape=[..., n, n]
            H_0 matrix.
        mat : array-like, shape=[..., n, n]
            Matrix such that its exponential is Cor^+.
        """
        if base_point is None:
            sym_mat = image_point
            mat = sym_mat + self.unique_diag_mat(sym_mat)
        else:
            mat = logmh(base_point)

        eigvals, eigvecs = gs.linalg.eigh(mat)

        h0_mat = gs.zeros(mat.shape)

        divided_diffs = self._divided_difference_exp(eigvals)

        n = h0_mat.shape[-1]
        for index_i in range(n):
            for index_j in range(n):
                val = 0
                for index_k in range(n):
                    for index_l in range(n):
                        val += (
                            eigvecs[..., index_i, index_k]
                            * eigvecs[..., index_j, index_k]
                            * eigvecs[..., index_i, index_l]
                            * eigvecs[..., index_j, index_l]
                            * divided_diffs[..., index_k, index_l]
                        )
                h0_mat[..., index_i, index_j] = val

        return h0_mat, mat

    def _tangent_diag_map(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Tangent unique diagonal matrix at image point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image of the base point.
        mat : array-like, shape=[..., n, n]
            Matrix such that its exponential is Cor^+.
        """
        h0_mat, mat = self._build_tangent_diag_aux_mat(
            image_point=image_point, base_point=base_point
        )
        e = gs.ones(h0_mat.shape[-1])
        vec = gs.matvec(
            gs.linalg.inv(h0_mat),
            gs.matvec(
                Matrices.to_diagonal(
                    SymMatrixLog.inverse_tangent(
                        image_point=mat, image_tangent_vec=image_tangent_vec
                    )
                ),
                e,
            ),
        )
        return gs.vec_to_diag(-vec), mat

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., n, n]
            Image point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        diff_D, sym_mat = self._tangent_diag_map(
            image_point=image_point,
            base_point=base_point,
            image_tangent_vec=image_tangent_vec,
        )
        return SymMatrixLog.inverse_tangent(
            image_point=sym_mat, image_tangent_vec=image_tangent_vec + diff_D
        )


class OffLogMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism between full-rank correlation matrices and
    hollow matrices endowed with a permutation-invariant metric.

    For more details, check section 8.2.2 [T2022]_.

    Parameters
    ----------
    space : FullRankCorrelationMatrices
    alpha : float
        Scalar multiplying first term of quadratic form.
    beta : float
        Scalar multiplying second term of quadratic form.
    gamma : float
        Scalar multiplying third term of quadratic form.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space, alpha=None, beta=None, gamma=1.0):
        diffeo = OffLogDiffeo()

        image_space = SymmetricHollowMatrices(n=space.n, equip=False).equip_with_metric(
            HollowMatricesPermutationInvariantMetric,
            alpha=alpha,
            beta=beta,
            gamma=gamma,
        )

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


class SPDScalingFinder:
    r"""Find unique positive diagonal matrix corresponding to an SPD matrix.

    That is, for every symmetric positive-definite matrix :math:`\Sigma`,
    there exists a unique positive diagonal matrix :math:`\Delta` such that
    :math:`\Delta \Sigma \Delta` is a symmetric positive-definite matrix
    with unit row sums.

    This result is known as the existence and uniqueness of the scaling of
    SPD matrices ([T2024]_, [MO1968]_, [JR2009]_).

    It finds the roots of the gradient of the strictly convex map

    .. math::

        F(D) = \frac{1}{2} \mathbb{1}^{\top} D^{\top}
        \Sigma D \mathbb{1}-\operatorname{tr}(\log (D))

    For details, check out [T2024]_'s section 3.5.


    Parameters
    ----------
    root_finder : RootFinder

    References
    ----------
    .. [T2024] Thanwerdas, Yann. “Permutation-Invariant Log-Euclidean Geometries
        on Full-Rank Correlation Matrices.”
        SIAM Journal on Matrix Analysis and Applications, 2024, 930–53.
        https://doi.org/10.1137/22M1538144.
    .. [MO1968] Marshall, Albert W., and Ingram Olkin.
        “Scaling of Matrices to Achieve Specified Row and Column Sums.”
        Numerische Mathematik 12, no. 1 (August 1, 1968): 83–90.
        https://doi.org/10.1007/BF02170999.
    .. [JR2009] Johnson, Charles R., and Robert Reams.
        “Scaling of Symmetric Matrices by Positive Diagonal Congruence.”
        Linear and Multilinear Algebra 57, no. 2 (March 1, 2009): 123–40.
        https://doi.org/10.1080/03081080600872327.
    """

    def __init__(self, root_finder=None):
        if root_finder is None:
            root_finder = NewtonMethod()

        self.root_finder = root_finder

    def _jacobian_f(self, spd_matrix, diag_vec):
        r"""Jacobian of objective function.

        .. math::
            J_{\Sigma}(D) = \Sigma D e - D^{-1} e

        Parameters
        ----------
        spd_matrix : array-like, shape=[..., n, n]
            Symmetric positive-definite matrix.
        diag_vec : array-like, shape=[..., n]
            Vector corresponding to the diagonal of a matrix.

        Returns
        -------
        jacobian : array-like, shape=[..., n]
        """
        return gs.matvec(spd_matrix, diag_vec) - 1.0 / diag_vec

    def _hessian_f(self, spd_matrix, diag_vec):
        r"""Hessian of objective function.

        .. math::
            H_{\Sigma}(D) = \Sigma + D^-2

        Parameters
        ----------
        spd_matrix : array-like, shape=[..., n, n]
            Symmetric positive-definite matrix.
        diag_vec : array-like, shape=[..., n]
            Vector corresponding to the diagonal of a matrix.

        Returns
        -------
        hessian : array-like, shape=[..., n, n]
        """
        return spd_matrix + gs.vec_to_diag(1.0 / diag_vec**2)

    def _call_single(self, spd_matrix):
        """Apply root finder to find scaling.

        Parameters
        ----------
        spd_matrix : array-like, shape=[n, n]
            Symmetric positive-definite matrix.

        Returns
        -------
        diag_vec : array-like, shape=[n]
            Scaling of spd_matrix.
        """
        x0 = gs.ones(spd_matrix.shape[-1])

        func = lambda x: self._jacobian_f(spd_matrix, x)
        jac = lambda x: self._hessian_f(spd_matrix, x)
        res = self.root_finder.root(func, x0, fun_jac=jac)
        return res.x

    def __call__(self, spd_matrix):
        """Apply Newton method to find scaling.

        Parameters
        ----------
        sym_mat : array-like, shape=[..., n, n]
            Symmetric positive-definite matrix.

        Returns
        -------
        diag_vec : array-like, shape=[..., n]
            Scaling of spd_matrix.
        """
        if spd_matrix.ndim == 2:
            return self._call_single(spd_matrix)

        batch_shape = spd_matrix.shape[:-2]
        if len(batch_shape) == 1:
            return gs.stack([self._call_single(sym_mat_) for sym_mat_ in spd_matrix])

        mat_shape = spd_matrix.shape[-2:]
        flat_spd_mat = gs.reshape(spd_matrix, (-1,) + mat_shape)
        out = gs.stack([self._call_single(sym_mat_) for sym_mat_ in flat_spd_mat])
        return gs.reshape(out, batch_shape + (mat_shape.shape[-1],))


class LogScalingDiffeo(Diffeo):
    r"""Off-log diffeomorphism from Cor+ to Row_1^+.

    A diffeomorphism between full-rank correlation matrices :math:`Cor+(n)`
    and the space of symmetric matrices with null row sums :math:`Row_0(n)`.

    .. math::

        \operatorname{Log} ^{\star} =
        \log \left(\mathcal{D}^{\star}(\Sigma)
        \star \Sigma\right): \operatorname{Cor}^{+}(n)
        \longrightarrow \operatorname{Row}_0(n)

    Check out [T2024]_ for more details.

    References
    ----------
    .. [T2024] Thanwerdas, Yann. “Permutation-Invariant Log-Euclidean Geometries
        on Full-Rank Correlation Matrices.”
        SIAM Journal on Matrix Analysis and Applications, 2024, 930–53.
        https://doi.org/10.1137/22M1538144.
    """

    def __init__(self):
        super().__init__()
        self.unique_diag_mat = SPDScalingFinder()

    def __call__(self, base_point):
        """Diffeomorphism at base point.

        NB: congruence action is implictly performed.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., n, n]
            Image point.
        """
        diag_vec = self.unique_diag_mat(base_point)
        unit_row_sum_spd = FullRankCorrelationMatrices.diag_action(diag_vec, base_point)

        return logmh(unit_row_sum_spd)

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., n, n]
            Base point.
        """
        return corr_map(expmh(image_point))

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image of the base point.
        """
        if image_point is None:
            diag_vec = self.unique_diag_mat(base_point)
            base_point_row_1 = FullRankCorrelationMatrices.diag_action(
                diag_vec, base_point
            )
        else:
            base_point_row_1 = expmh(image_point)

        delta = Matrices.diagonal(base_point_row_1) ** 0.5
        aux = FullRankCorrelationMatrices.diag_action(delta, tangent_vec)
        eye = gs.eye(base_point_row_1.shape[-1])
        tangent_vec_row_1_0 = -2 * gs.vec_to_diag(
            gs.sum(
                Matrices.mul(
                    gs.linalg.inv(eye + base_point_row_1),
                    aux,
                ),
                axis=-1,
            )
        )

        tangent_vec_row_1 = aux + 0.5 * (
            Matrices.mul(tangent_vec_row_1_0, base_point_row_1)
            + Matrices.mul(base_point_row_1, tangent_vec_row_1_0)
        )

        return SymMatrixLog.tangent(
            tangent_vec=tangent_vec_row_1,
            base_point=base_point_row_1,
            image_point=image_point,
        )

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., n, n]
            Image point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        if image_point is not None:
            base_point_row_1 = expmh(image_point)
        else:
            diag_vec = self.unique_diag_mat(base_point)
            base_point_row_1 = FullRankCorrelationMatrices.diag_action(
                diag_vec, base_point
            )

        tangent_vec_row_1 = SymMatrixLog.inverse_tangent(
            image_tangent_vec=image_tangent_vec,
            image_point=image_point,
            base_point=base_point_row_1,
        )
        return tangent_corr_map(tangent_vec_row_1, base_point_row_1)


class LogScaledMetric(PullbackDiffeoMetric):
    """Pullback metric via a diffeomorphism.

    Diffeormorphism between full-rank correlation matrices and
    the space of symmetric matrices with null row sums.

    Check out [T2024]_ for more details.

    Parameters
    ----------
    space : FullRankCorrelationMatrices
    alpha : float
        Scalar multiplying first term of quadratic form.
    delta : float
        Scalar multiplying second term of quadratic form.
    zeta : float
        Scalar multiplying third term of quadratic form.

    References
    ----------
    .. [T2024] Thanwerdas, Yann. “Permutation-Invariant Log-Euclidean Geometries
        on Full-Rank Correlation Matrices.”
        SIAM Journal on Matrix Analysis and Applications, 2024, 930–53.
        https://doi.org/10.1137/22M1538144.
    """

    def __init__(self, space, alpha=None, delta=None, zeta=1.0):
        diffeo = LogScalingDiffeo()

        image_space = NullRowSumsSymmetricMatrices(
            n=space.n, equip=False
        ).equip_with_metric(
            NullRowSumsPermutationInvariantMetric,
            alpha=alpha,
            delta=delta,
            zeta=zeta,
        )

        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


register_quotient(
    Space=SPDMatrices,
    Metric=SPDAffineMetric,
    GroupAction=FullRankCorrelationMatrices.diag_action,
    FiberBundle=CorrelationMatricesBundle,
)



"""Exponential barycenter.

Lead author: Nicolas Guigui.
"""

import logging

from sklearn.base import BaseEstimator

import geomstats.backend as gs
from geomstats.geometry.euclidean import Euclidean
from geomstats.learning.frechet_mean import BaseGradientDescent, LinearMean


class GradientDescent(BaseGradientDescent):
    """Gradient descent for exponential barycenter."""

    def minimize(self, group, points, weights=None):
        """Compute the (weighted) group exponential barycenter of `points`.

        Parameters
        ----------
        group : LieGroup
            Instance of the class LieGroup.
        points : array-like, shape=[n_samples, dim, dim]
            Input points lying in the Lie Group.
        weights : array-like, shape=[n_samples,]
            Weights associated to the points.
            Optional, defaults to 1 for each point if None.

        Returns
        -------
        exp_bar : array-like, shape=[dim, dim]
            Exponential barycenter of the input points.
        """
        n_samples = points.shape[0]

        if n_samples == 1:
            return points[0]

        if weights is None:
            weights = gs.ones((n_samples,))

        sum_weights = gs.sum(weights)

        mean = points[0] if self.init_point is None else self.init_point

        grad_norm = 0.0

        for iteration in range(self.max_iter):
            if not (grad_norm > self.epsilon or iteration == 0):
                break

            inv_mean = group.inverse(mean)
            centered_points = group.compose(inv_mean, points)

            logs = group.log(point=centered_points)
            tangent_mean = self.init_step_size * gs.einsum(
                "n, nk...->k...", weights / sum_weights, logs
            )
            mean_next = group.compose(mean, group.exp(tangent_vec=tangent_mean))

            grad_norm = gs.linalg.norm(tangent_mean)

            mean = mean_next

        else:
            logging.warning(
                f"Maximum number of iterations {self.max_iter} reached. "
                "The mean may be inaccurate"
            )

        if self.verbose:
            logging.info(f"n_iter: {iteration}, final gradient norm: {grad_norm}")
        return mean


class ExponentialBarycenter(BaseEstimator):
    """Empirical exponential barycenter for matrix groups.

    Parameters
    ----------
    space : LieGroup
        Lie group instance on which the data lie.

    Attributes
    ----------
    estimate_ : array-like, shape=[dim, dim]
        If fit, exponential barycenter.
    """

    def __new__(cls, space):
        """Interface for instantiating proper algorithm."""
        if isinstance(space, Euclidean):
            return LinearMean(space)

        return super().__new__(cls)

    def __init__(self, space):
        self.space = space

        self.optimizer = GradientDescent()

        self.estimate_ = None

    def set(self, **kwargs):
        """Set optimizer parameters.

        Especially useful for one line instantiations.
        """
        for param_name, value in kwargs.items():
            if not hasattr(self.optimizer, param_name):
                raise ValueError(f"Unknown parameter {param_name}.")

            setattr(self.optimizer, param_name, value)
        return self

    def fit(self, X, y=None, weights=None):
        """Compute the empirical weighted exponential barycenter.

        Parameters
        ----------
        X : array-like, shape=[n_samples, dim, dim]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        # TODO (nguigs): use closed form expression for special euclidean
        #  group as before PR #537

        self.estimate_ = self.optimizer.minimize(
            group=self.space,
            points=X,
            weights=weights,
        )

        return self



from geomstats.geometry.euclidean import Euclidean


class EuclideanGroup(Euclidean):
    @staticmethod
    def compose(point_a, point_b):
        return point_a + point_b

    def log(self, point, base_point=None):
        if base_point is None:
            base_point = self.identity

        return point - base_point

    def exp(self, tangent_vec, base_point=None):
        if base_point is None:
            return tangent_vec

        return super().exp(tangent_vec, base_point)

    @staticmethod
    def inverse(point):
        return -point



"""Product of manifolds.

Lead author: Nicolas Guigui, John Harvey.
"""

import math

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.complex_manifold import ComplexManifold
from geomstats.geometry.complex_riemannian_metric import ComplexRiemannianMetric
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import get_batch_shape

COMPLEX_OBJECTS = (ComplexRiemannianMetric, ComplexManifold)


def _factor_is_complex(factor):
    if (
        isinstance(factor, COMPLEX_OBJECTS)
        or hasattr(factor, "underlying_metric")
        and isinstance(factor.underlying_metric, COMPLEX_OBJECTS)
    ):
        return True

    return False


def _has_mixed_fields(factors):
    bools = [_factor_is_complex(factor) for factor in factors]
    if len(set(bools)) == 2:
        return True

    return False


def _all_equal(arg):
    """Check if all elements of arg are equal."""
    return arg.count(arg[0]) == len(arg)


def _block_diagonal(factor_matrices):
    """Put a list of square matrices in block diagonal form."""
    shapes_dict = {}
    for i, matrix_i in enumerate(factor_matrices):
        for j, matrix_j in enumerate(factor_matrices):
            shapes_dict[(i, j)] = matrix_i.shape[:-1] + matrix_j.shape[-1:]
    rows = []
    # concacatenate along axis = -2
    for i, matrix_i in enumerate(factor_matrices):
        # concatenate along axis = -1
        blocks_to_concatenate = []
        for j, _ in enumerate(factor_matrices):
            if i == j:
                blocks_to_concatenate.append(matrix_i)
            else:
                blocks_to_concatenate.append(gs.zeros(shapes_dict[(i, j)]))
        row = gs.concatenate(blocks_to_concatenate, axis=-1)
        rows.append(row)
    metric_matrix = gs.concatenate(rows, axis=-2)
    return metric_matrix


def _find_product_shape(factors, point_ndim):
    """Determine an appropriate shape for the product from the factors."""
    factor_shapes = [factor.shape for factor in factors]

    if point_ndim is None:
        if _all_equal(factor_shapes):
            return len(factors), *factors[0].shape
        point_ndim = 1

    if point_ndim == 1:
        return (sum(math.prod(factor_shape) for factor_shape in factor_shapes),)
    if not _all_equal(factor_shapes):
        raise ValueError(
            "A point_ndim greater than one can only be used if all "
            "manifolds have shape."
        )
    return len(factors), *factors[0].shape


class _IterateOverFactorsMixins:
    def __init__(
        self, factors, cum_index, pool_outputs, has_mixed_fields, *args, **kwargs
    ):
        self.factors = factors
        self._cum_index = cum_index
        self._pool_outputs = pool_outputs
        self._has_mixed_fields = has_mixed_fields
        super().__init__(*args, **kwargs)

    def embed_to_product(self, points):
        """Map a point in each factor to a point in the product.

        Parameters
        ----------
        points : list
            A list of points, one from each factor, each array-like of shape
            (..., factor.shape)

        Returns
        -------
        point : array-like, shape (..., shape)

        Raises
        ------
        ShapeError
            If the points are not compatible with the shapes of the corresponding
            factors.
        """
        for point, factor in zip(points, self.factors):
            geomstats.errors.check_point_shape(point, factor)

        if self.point_ndim == 1:
            points_ = []
            for point, factor in zip(points, self.factors):
                if gs.ndim(point) > len(factor.shape):
                    batch_shape = get_batch_shape(factor.point_ndim, point)
                    point = gs.reshape(point, batch_shape + (-1,))
                else:
                    point = gs.flatten(point)

                points_.append(point)
            return gs.concatenate(points_, axis=-1)
        stacking_axis = -1 * len(self.shape)
        return gs.stack(points, axis=stacking_axis)

    def project_from_product(self, point):
        """Map a point in the product to points in each factor.

        Parameters
        ----------
        point : array-like, shape (..., shape)
            The point to be projected to the factors

        Returns
        -------
        projected_points : list of array-like
            The points on each factor, of shape (..., factor.shape)

        Raises
        ------
        ShapeError
            If the point does not have a shape compatible with the product manifold.
        """
        geomstats.errors.check_point_shape(point, self)

        if self.point_ndim == 1:
            projected_points = gs.split(point, self._cum_index, axis=-1)
            projected_points = [
                self._reshape_trailing(projected_points[j], self.factors[j])
                for j in range(len(self.factors))
            ]

        else:
            splitting_axis = -1 * len(self.shape)
            projected_points = gs.split(point, len(self.factors), axis=splitting_axis)
            projected_points = [
                gs.squeeze(projected_point, axis=splitting_axis)
                for projected_point in projected_points
            ]

        if self._has_mixed_fields:
            for i, (factor, projected_point) in enumerate(
                zip(self.factors, projected_points)
            ):
                if not _factor_is_complex(factor):
                    projected_points[i] = gs.real(projected_point)

        return projected_points

    @staticmethod
    def _reshape_trailing(argument, factor):
        """Convert the trailing dimensions to match the shape of a factor manifold."""
        space = factor._space if isinstance(factor, RiemannianMetric) else factor

        if space.point_ndim == 1:
            return argument
        leading_shape = argument.shape[:-1]
        trailing_shape = space.shape
        new_shape = leading_shape + trailing_shape
        return gs.reshape(argument, new_shape)

    def _iterate_over_factors(self, func, args):
        """Apply a function to each factor of the product.

        func is called on each factor of the product.

        Array-type arguments are separated out to be passed to func for each factor,
        but other arguments are passed unchanged.

        Parameters
        ----------
        func : str
            The name of a method which is defined for each factor of the product
            The method must return an array of shape (..., factor.shape) or a boolean
            array of shape (...,).
        args : dict
            Dict of arguments.
            Array-type arguments must be of type (..., shape)
            Other arguments are passed to each factor unchanged

        Returns
        -------
        out : array-like, shape = [..., {(), shape}]
        """
        # TODO The user may prefer to provide the arguments as lists and receive them as
        # TODO lists, as this may be the form in which they are available. This should
        # TODO be allowed, rather than packing and unpacking them repeatedly.
        args_list, numerical_args = self._validate_and_prepare_args_for_iteration(args)

        out = [
            self._get_method(self.factors[i], func, args_list[i], numerical_args)
            for i in range(len(self.factors))
        ]
        if self._pool_outputs:
            return self._pool_outputs_from_function(out)
        return out

    def _validate_and_prepare_args_for_iteration(self, args):
        """Separate arguments into different types and validate them.

        Parameters
        ----------
        args : dict
            Dict of arguments.
            Float or int arguments are passed to func for each manifold
            Array-type arguments must be of type (..., shape)

        Returns
        -------
        arguments : list
            List of dicts of arguments with values being array-like.
            Each element of the list corresponds to a factor af the manifold.
        numerical_args : dict
            Dict of non-array arguments
        """
        args_list = [{} for _ in self.factors]
        numerical_args = {}
        for key, value in args.items():
            if not gs.is_array(value):
                numerical_args[key] = value
            else:
                new_args = self.project_from_product(value)
                for args_dict, new_arg in zip(args_list, new_args):
                    args_dict[key] = new_arg
        return args_list, numerical_args

    @staticmethod
    def _get_method(factor, method_name, array_args, num_args):
        """Call factor.method_name."""
        return getattr(factor, method_name)(**array_args, **num_args)


class ProductManifold(_IterateOverFactorsMixins, Manifold):
    """Class for a product of manifolds M_1 x ... x M_n.

    In contrast to the classes NFoldManifold, Landmarks, or DiscretizedCurves,
    the manifolds M_1, ..., M_n need not be the same, nor of
    same dimension, but the list of manifolds needs to be provided.

    Parameters
    ----------
    factors : tuple
        Collection of manifolds in the product.
    point_ndim : int or None
        If None, defaults to 1, unless all factors have the same shape.
    """

    def __init__(self, factors, point_ndim=None, equip=True):
        factors = tuple(factors)

        factor_dims = [factor.dim for factor in factors]

        dim = sum(factor_dims)

        shape = _find_product_shape(factors, point_ndim)

        intrinsic = all(factor.intrinsic for factor in factors)

        if not intrinsic:
            factor_embedding_spaces = [
                (
                    manifold.embedding_space
                    if hasattr(manifold, "embedding_space")
                    else manifold
                )
                for manifold in factors
            ]
            # TODO: need to revisit due to removal of scales
            self.embedding_space = ProductManifold(
                factor_embedding_spaces, point_ndim, equip=False
            )

        cum_index = (
            gs.cumsum(factor_dims)[:-1]
            if intrinsic
            else self.embedding_space._cum_index
        )

        super().__init__(
            factors=factors,
            cum_index=cum_index,
            pool_outputs=True,
            has_mixed_fields=_has_mixed_fields(factors),
            dim=dim,
            shape=shape,
            intrinsic=intrinsic,
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ProductRiemannianMetric

    def _pool_outputs_from_function(self, outputs):
        """Collect outputs for each product to be returned.

        If each element of the output is a boolean array of the same shape, test along
        the list whether all elements are True and return a boolean array of the same
        shape.

        Otherwise, if each element of the output has a shape compatible with points of
        the corresponding factor, an attempt is made to map the list of points to a
        point in the product by embed_to_product.

        Parameters
        ----------
        outputs : list
            A list of outputs which must be pooled

        Returns
        -------
        pooled_output : array-like, shape {(...,), (..., self.shape)}
        """
        # TODO: simplify after cleaning gs.squeeze
        all_arrays = gs.all([gs.is_array(factor_output) for factor_output in outputs])
        if (
            all_arrays
            and _all_equal([factor_output.shape for factor_output in outputs])
            and gs.all([gs.is_bool(factor_output) for factor_output in outputs])
            or (not all_arrays)
        ):
            outputs = gs.stack([gs.array(factor_output) for factor_output in outputs])
            outputs = gs.all(outputs, axis=0)
            return outputs

        try:
            return self.embed_to_product(outputs)
        except geomstats.errors.ShapeError:
            raise RuntimeError(
                "Could not combine outputs - they are not points of the individual"
                " factors."
            )
        except ValueError:
            raise RuntimeError(
                "Could not combine outputs, probably because they could"
                " not be concatenated or stacked."
            )

    def belongs(self, point, atol=gs.atol):
        """Test if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Point.
        atol : float,
            Tolerance.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if the point belongs to the manifold.
        """
        belongs = self._iterate_over_factors("belongs", {"point": point, "atol": atol})
        return belongs

    def regularize(self, point):
        """Regularize the point into the manifold's canonical representation.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Point to be regularized.

        Returns
        -------
        regularized_point : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Point in the manifold's canonical representation.
        """
        regularized_point = self._iterate_over_factors("regularize", {"point": point})
        return regularized_point

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in the product space from the product distribution.

        The distribution used is the product of the distributions used by the
        random_point methods of each individual factor manifold.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Points sampled from the manifold.
        """
        samples = self._iterate_over_factors(
            "random_point", {"n_samples": n_samples, "bound": bound}
        )
        return samples

    def random_tangent_vec(self, base_point, n_samples=1):
        """Sample on the tangent space from the product distribution.

        The distribution used is the product of the distributions used by the
        random_tangent_vec methods of each individual factor manifold.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point of the tangent space.
            Optional, default: None.
        n_samples : int
            Number of samples.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Points sampled in the tangent space of the product manifold at base_point.
        """
        samples = self._iterate_over_factors(
            "random_tangent_vec", {"base_point": base_point, "n_samples": n_samples}
        )
        return samples

    def projection(self, point):
        """Project a point onto product manifold.

        Parameters
        ----------
        point : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Point in product manifold.

        Returns
        -------
        projected : array-like, shape=[..., {dim, embedding_space.dim, \
            [n_manifolds, dim_each]}]
            Projected point.
        """
        projected_point = self._iterate_over_factors("projection", {"point": point})
        return projected_point

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.

        Notes
        -----
        The tangent space of the product manifold is the direct sum of
        tangent spaces.
        """
        tangent_vec = self._iterate_over_factors(
            "to_tangent", {"base_point": base_point, "vector": vector}
        )
        return tangent_vec

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        The tangent space of the product manifold is the direct sum of
        tangent spaces.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.
            Optional, default: None
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        is_tangent = self._iterate_over_factors(
            "is_tangent", {"base_point": base_point, "vector": vector, "atol": atol}
        )
        return is_tangent


class ProductRiemannianMetric(_IterateOverFactorsMixins, RiemannianMetric):
    """Class for product of Riemannian metrics."""

    def __init__(self, space):
        factors = [factor.metric for factor in space.factors]
        factor_signatures = [metric.signature for metric in factors]

        sig_pos = sum(sig[0] for sig in factor_signatures)
        sig_neg = sum(sig[1] for sig in factor_signatures)

        super().__init__(
            space=space,
            factors=factors,
            cum_index=space._cum_index,
            pool_outputs=False,
            has_mixed_fields=space._has_mixed_fields,
            signature=(sig_pos, sig_neg),
        )

    @property
    def shape(self):
        """Shape of space."""
        return self._space.shape

    @property
    def point_ndim(self):
        """Point type of space."""
        return self._space.point_ndim

    def metric_matrix(self, base_point=None):
        """Compute the matrix of the inner-product.

        Matrix of the inner-product defined by the Riemmanian metric
        at point base_point of the manifold.

        Parameters
        ----------
        base_point : array-like, shape=[..., self.shape]
            Point on the manifold at which to compute the inner-product matrix.
            Optional, default: None.

        Returns
        -------
        matrix : array-like, shape as described below
            Matrix of the inner-product at the base point.
            The matrix is in block diagonal form with a block for each factor.
            Each block is the same size as the metric_matrix for that factor.
        """
        factor_matrices = self._iterate_over_factors(
            "metric_matrix", {"base_point": base_point}
        )
        return _block_diagonal(factor_matrices)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the inner-product of two tangent vectors at a base point.

        Inner product defined by the Riemannian metric at point `base_point`
        between tangent vectors `tangent_vec_a` and `tangent_vec_b`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., self.shape]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., self.shape]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., self.shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Inner-product of the two tangent vectors.
        """
        args = {
            "tangent_vec_a": tangent_vec_a,
            "tangent_vec_b": tangent_vec_b,
            "base_point": base_point,
        }
        inner_products = self._iterate_over_factors("inner_product", args)
        return sum(inner_products)

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., self.shape]
            Vector.
        base_point : array-like, shape=[..., self.shape]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        args = {
            "vector": vector,
            "base_point": base_point,
        }
        sq_norms = self._iterate_over_factors("squared_norm", args)
        return sum(sq_norms)

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a tangent vector.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., self.shape]
            Tangent vector at a base point.
        base_point : array-like, shape=[..., self.shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        exp : array-like, shape=[..., self.shape]
            Point on the manifold equal to the Riemannian exponential
            of tangent_vec at the base point.
        """
        args = {"tangent_vec": tangent_vec, "base_point": base_point}
        exp = self._iterate_over_factors("exp", args)
        return self._space.embed_to_product(exp)

    def log(self, point, base_point):
        """Compute the Riemannian logarithm of a point.

        Parameters
        ----------
        point : array-like, shape=[..., self.shape]
            Point on the manifold.
        base_point : array-like, shape=[..., self.shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        log : array-like, shape=[..., self.shape]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        args = {"point": point, "base_point": base_point}
        logs = self._iterate_over_factors("log", args)
        return self._space.embed_to_product(logs)

    def dist(self, point_a, point_b):
        """Geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., self.shape]
            Point.
        point_b : array-like, shape=[..., self.shape]
            Point.

        Returns
        -------
        dist : array-like, shape=[...,]
            Distance.
        """
        args = {"point_a": point_a, "point_b": point_b}
        dists = gs.array(self._iterate_over_factors("dist", args))
        return gs.linalg.norm(dists, ord=2, axis=0)

    def geodesic(self, initial_point, end_point=None, initial_tangent_vec=None):
        """Generate parameterized function for the geodesic curve.

        Geodesic curve defined by either:

        - an initial point and an initial tangent vector,
        - an initial point and an end point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Point on the manifold, initial point of the geodesic.
        end_point : array-like, shape=[..., dim], optional
            Point on the manifold, end point of the geodesic. If None,
            an initial tangent vector must be given.
        initial_tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point, the initial speed of the geodesics.
            Optional, default: None.
            If None, an end point must be given and a logarithm is computed.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve. If a batch of initial
            conditions is passed, the output array's first dimension
            represents the different initial conditions, and the second
            corresponds to time.
        """
        args = {
            "initial_point": initial_point,
            "end_point": end_point,
            "initial_tangent_vec": initial_tangent_vec,
        }
        geodesics = self._iterate_over_factors("geodesic", args)

        def geod_fun(t):
            t = gs.to_ndarray(t, to_ndim=1)
            values = [geodesic(t) for geodesic in geodesics]
            return self._space.embed_to_product(values)

        return geod_fun



import pytest

import geomstats.backend as gs
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class PointToGeodesicAlignerTestCase(TestCase):
    def test_align(self, geodesic, point, expected, atol):
        res = self.aligner.align(geodesic, point)
        self.assertAllClose(res, expected, atol=atol)

    def test_align_with_endpoints(
        self, initial_point, end_point, point, expected, atol
    ):
        geodesic = self.total_space.quotient.metric.geodesic(initial_point, end_point)
        self.test_align(geodesic, point, expected, atol)

    @pytest.mark.vec
    def test_align_vec(self, n_reps, atol):
        initial_point = self.data_generator.random_point()
        end_point = self.data_generator.random_point()

        geodesic = self.total_space.quotient.metric.geodesic(initial_point, end_point)
        point = self.data_generator.random_point()

        expected = self.aligner.align(geodesic, point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    initial_point=initial_point,
                    end_point=end_point,
                    point=point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["initial_point", "end_point", "point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data, test_fnc_name="test_align_with_endpoints")

    def test_dist(self, geodesic, point, expected, atol):
        res = self.aligner.dist(geodesic, point)
        self.assertAllClose(res, expected, atol=atol)

    def test_dist_with_endpoints(self, initial_point, end_point, point, expected, atol):
        geodesic = self.total_space.quotient.metric.geodesic(initial_point, end_point)
        self.test_dist(geodesic, point, expected, atol)

    @pytest.mark.vec
    def test_dist_vec(self, n_reps, atol):
        initial_point = self.data_generator.random_point()
        end_point = self.data_generator.random_point()

        geodesic = self.total_space.quotient.metric.geodesic(initial_point, end_point)
        point = self.data_generator.random_point()

        expected = self.aligner.dist(geodesic, point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    initial_point=initial_point,
                    end_point=end_point,
                    point=point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["initial_point", "end_point", "point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data, test_fnc_name="test_dist_with_endpoints")

    @pytest.mark.random
    def test_dist_along_geodesic_is_zero(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        geodesic = self.total_space.quotient.metric.geodesic(
            initial_point,
            end_point,
        )

        s = gs.random.rand(1)
        points = gs.squeeze(geodesic(s), axis=-3)

        batch_shape = (n_points,) if n_points > 1 else ()
        self.test_dist_with_endpoints(
            initial_point, end_point, points, gs.zeros(batch_shape), atol
        )



"""Numpy based computation backend."""

import numpy as _np
from numpy import (
    all,
    allclose,
    amax,
    amin,
    any,
    argmax,
    argmin,
    broadcast_arrays,
    broadcast_to,
    clip,
    complex64,
    complex128,
    concatenate,
    conj,
    cross,
    cumprod,
    cumsum,
    diag_indices,
    diagonal,
    einsum,
    empty_like,
    equal,
    expand_dims,
    flip,
    float32,
    float64,
    greater,
    hsplit,
    hstack,
    int32,
    int64,
    isclose,
    isnan,
    kron,
    less,
    less_equal,
    logical_and,
    logical_or,
    maximum,
    mean,
    meshgrid,
    minimum,
    moveaxis,
    ones_like,
    pad,
    prod,
    quantile,
    repeat,
    reshape,
    searchsorted,
    shape,
    sort,
    split,
    stack,
    std,
    sum,
    take,
    tile,
    transpose,
    trapz,
    tril,
    tril_indices,
    triu,
    triu_indices,
    uint8,
    unique,
    vstack,
    where,
    zeros_like,
)
from scipy.special import erf, gamma, polygamma  # NOQA

from .._shared_numpy import (
    abs,
    angle,
    arange,
    arccos,
    arccosh,
    arcsin,
    arctan2,
    arctanh,
    array_from_sparse,
    assignment,
    assignment_by_sum,
    ceil,
    copy,
    cos,
    cosh,
    divide,
    dot,
    exp,
    flatten,
    floor,
    from_numpy,
    get_slice,
    imag,
    log,
    mat_from_diag_triu_tril,
    matmul,
    matvec,
    mod,
    ndim,
    one_hot,
    outer,
    power,
    ravel_tril_indices,
    real,
    scatter_add,
    set_diag,
    sign,
    sin,
    sinh,
    sqrt,
    squeeze,
    tan,
    tanh,
    to_numpy,
    trace,
    tril_to_vec,
    triu_to_vec,
    vec_to_diag,
    vectorize,
)
from . import (
    autodiff,  # NOQA
    linalg,  # NOQA
    random,  # NOQA
)
from ._common import (
    _box_binary_scalar,
    _box_unary_scalar,
    _dyn_update_dtype,
    _modify_func_default_dtype,
    array,
    as_dtype,
    atol,
    cast,
    convert_to_wider_dtype,
    eye,
    get_default_cdtype,
    get_default_dtype,
    is_array,
    is_bool,
    is_complex,
    is_floating,
    rtol,
    set_default_dtype,
    to_ndarray,
    zeros,
)

ones = _modify_func_default_dtype(target=_np.ones)
linspace = _dyn_update_dtype(target=_np.linspace, dtype_pos=5)
empty = _dyn_update_dtype(target=_np.empty, dtype_pos=1)


def has_autodiff():
    """If allows for automatic differentiation.

    Returns
    -------
    has_autodiff : bool
    """
    return False



import math

import pytest

import geomstats.backend as gs
from geomstats.geometry.spd_matrices import SPDMatrices
from geomstats.test_cases.geometry.connection import (
    ConnectionComparisonTestCase,
    ConnectionTestCase,
)
from geomstats.test_cases.geometry.mixins import DistTestCaseMixins
from geomstats.vectorization import get_batch_shape


class RiemannianMetricTestCase(DistTestCaseMixins, ConnectionTestCase):
    def test_metric_matrix(self, base_point, expected, atol):
        res = self.space.metric.metric_matrix(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_metric_matrix_is_spd(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        metric_matrix = self.space.metric.metric_matrix(base_point)

        res = SPDMatrices(n=math.prod(self.space.shape)).belongs(
            metric_matrix, atol=atol
        )
        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)

        self.assertAllEqual(res, expected)

    def test_cometric_matrix(self, base_point, expected, atol):
        res = self.space.metric.cometric_matrix(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_inner_product_derivative_matrix(self, base_point, expected, atol):
        res = self.space.metric.inner_product_derivative_matrix(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_inner_product(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        # TODO: test inner_product with itself?
        res = self.space.metric.inner_product(tangent_vec_a, tangent_vec_b, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_inner_product_is_symmetric(self, n_points, atol):
        """Check inner product is symmetric.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        inner_product_ab = self.space.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        inner_product_ba = self.space.metric.inner_product(
            tangent_vec_b, tangent_vec_a, base_point
        )

        self.assertAllClose(inner_product_ab, inner_product_ba, atol=atol)

    def test_inner_coproduct(
        self, cotangent_vec_a, cotangent_vec_b, base_point, expected, atol
    ):
        res = self.space.metric.inner_coproduct(
            cotangent_vec_a, cotangent_vec_b, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_hamiltonian(self, state, expected, atol):
        res = self.space.metric.hamiltonian(state)
        self.assertAllClose(res, expected, atol=atol)

    def test_squared_norm(self, vector, base_point, expected, atol):
        res = self.space.metric.squared_norm(vector, base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_norm(self, vector, base_point, expected, atol):
        res = self.space.metric.norm(vector, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_norm_is_positive(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        vector = self.data_generator.random_tangent_vec(base_point)

        norm_ = self.space.metric.norm(vector, base_point)

        res = gs.all(norm_ > -atol)
        self.assertTrue(res)

    def test_normalize(self, vector, base_point, expected, atol):
        res = self.space.metric.normalize(vector, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_normalize_is_unitary(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        vec = self.data_generator.random_tangent_vec(point)

        normalized_vec = self.space.metric.normalize(vec, point)
        res = self.space.metric.norm(normalized_vec, point)

        batch_shape = (n_points,) if n_points > 1 else ()
        expected = gs.ones(batch_shape)
        self.assertAllClose(res, expected, atol=atol)

    def test_squared_dist(self, point_a, point_b, expected, atol):
        res = self.space.metric.squared_dist(point_a, point_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_squared_dist_is_symmetric(self, n_points, atol):
        """Check squared distance is symmetric.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        squared_dist_ab = self.space.metric.squared_dist(point_a, point_b)
        squared_dist_ba = self.space.metric.squared_dist(point_b, point_a)

        self.assertAllClose(squared_dist_ab, squared_dist_ba, atol=atol)

    @pytest.mark.random
    def test_squared_dist_is_positive(self, n_points, atol):
        """Check squared distance is positive.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        squared_dist_ = self.space.metric.squared_dist(point_a, point_b)
        res = gs.all(squared_dist_ > -atol)
        self.assertTrue(res)

    @pytest.mark.random
    def test_dist_is_log_norm(self, n_points, atol):
        """Check distance is norm of log.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        log_norm = self.space.metric.norm(
            self.space.metric.log(point_b, point_a), point_a
        )
        dist_ = self.space.metric.dist(point_a, point_b)
        self.assertAllClose(dist_, log_norm, atol=atol)

    def test_diameter(self, points, expected, atol):
        res = self.space.metric.diameter(points)
        self.assertAllClose(res, expected, atol=atol)

    def test_normal_basis(self, basis, base_point, expected, atol):
        res = self.space.metric.normal_basis(basis, base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_covariant_riemann_tensor(self, base_point, expected, atol):
        res = self.space.metric.covariant_riemann_tensor(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_covariant_riemann_tensor_is_skew_symmetric_1(self, n_points, atol):
        """Check covariant riemannian tensor verifies first skew symmetry.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        # TODO: add definition of first skew symmetry in docstrings
        base_point = self.data_generator.random_point(n_points)

        covariant_metric_tensor = self.space.metric.covariant_riemann_tensor(base_point)
        skew_symmetry_1 = covariant_metric_tensor + gs.moveaxis(
            covariant_metric_tensor, [-2, -1], [-1, -2]
        )

        res = gs.all(gs.abs(skew_symmetry_1) < atol)
        self.assertTrue(res)

    @pytest.mark.random
    def test_covariant_riemann_tensor_is_skew_symmetric_2(self, n_points, atol):
        """Check covariant riemannian tensor verifies second skew symmetry.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        # TODO: add definition of second skew symmetry in docstrings
        base_point = self.data_generator.random_point(n_points)

        covariant_metric_tensor = self.space.metric.covariant_riemann_tensor(base_point)
        skew_symmetry_2 = covariant_metric_tensor + gs.moveaxis(
            covariant_metric_tensor, [-4, -3], [-3, -4]
        )

        res = gs.all(gs.abs(skew_symmetry_2) < atol)
        self.assertTrue(res)

    @pytest.mark.random
    def test_covariant_riemann_tensor_bianchi_identity(self, n_points, atol):
        """Check covariant riemannian tensor verifies Bianchi identity.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        # TODO: add Bianchi identity in docstrings
        base_point = self.data_generator.random_point(n_points)

        covariant_metric_tensor = self.space.metric.covariant_riemann_tensor(base_point)
        bianchi_identity = (
            covariant_metric_tensor
            + gs.moveaxis(covariant_metric_tensor, [-3, -2, -1], [-2, -1, -3])
            + gs.moveaxis(covariant_metric_tensor, [-3, -2, -1], [-1, -3, -2])
        )

        res = gs.all(gs.abs(bianchi_identity) < gs.atol)
        self.assertTrue(res)

    @pytest.mark.random
    def test_covariant_riemann_tensor_is_interchange_symmetric(self, n_points, atol):
        """Check covariant riemannian tensor verifies interchange symmetry.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        # TODO: add definition of interchange symmetry in docstrings
        base_point = self.data_generator.random_point(n_points)

        covariant_metric_tensor = self.space.metric.covariant_riemann_tensor(base_point)
        interchange_symmetry = covariant_metric_tensor - gs.moveaxis(
            covariant_metric_tensor, [-4, -3, -2, -1], [-2, -1, -4, -3]
        )

        res = gs.all(gs.abs(interchange_symmetry) < atol)
        self.assertTrue(res)

    def test_sectional_curvature(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_scalar_curvature(self, base_point, expected, atol):
        res = self.space.metric.scalar_curvature(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_parallel_transport_ivp_norm(self, n_points, atol):
        """Check parallel transported norm is preserved.

        This is for parallel transport defined by initial value problem (ivp).

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        direction = self.data_generator.random_tangent_vec(base_point)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        transported = self.space.metric.parallel_transport(
            tangent_vec, base_point, direction=direction
        )

        end_point = self.space.metric.exp(direction, base_point)

        self.assertAllClose(
            self.space.metric.norm(transported, end_point),
            self.space.metric.norm(tangent_vec, base_point),
            atol=atol,
        )

    @pytest.mark.random
    def test_parallel_transport_bvp_norm(self, n_points, atol):
        """Check parallel transported norm is preserved.

        This is for parallel transport defined by boundary value problem (bvp).

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        transported = self.space.metric.parallel_transport(
            tangent_vec, base_point, end_point=end_point
        )

        self.assertAllClose(
            self.space.metric.norm(transported, end_point),
            self.space.metric.norm(tangent_vec, base_point),
            atol=atol,
        )


class RiemannianMetricComparisonTestCase(ConnectionComparisonTestCase):
    def test_metric_matrix(self, base_point, atol):
        res = self.space.metric.metric_matrix(base_point)
        res_ = self.other_space.metric.metric_matrix(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_metric_matrix_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_metric_matrix(base_point, atol)

    def test_cometric_matrix(self, base_point, atol):
        res = self.space.metric.cometric_matrix(base_point)
        res_ = self.other_space.metric.cometric_matrix(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_cometric_matrix_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_cometric_matrix(base_point, atol)

    def test_inner_product_derivative_matrix(self, base_point, atol):
        res = self.space.metric.inner_product_derivative_matrix(base_point)
        res_ = self.other_space.metric.inner_product_derivative_matrix(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_inner_product_derivative_matrix_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_inner_product_derivative_matrix(base_point, atol)

    def test_inner_product(self, tangent_vec_a, tangent_vec_b, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        tangent_vec_a_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_a, base_point
        )
        tangent_vec_b_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_b, base_point
        )

        res = self.space.metric.inner_product(tangent_vec_a, tangent_vec_b, base_point)
        res_ = self.other_space.metric.inner_product(
            tangent_vec_a_, tangent_vec_b_, base_point_
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_inner_product_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        self.test_inner_product(tangent_vec_a, tangent_vec_b, base_point, atol)

    def test_inner_coproduct(self, cotangent_vec_a, cotangent_vec_b, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        cotangent_vec_a_ = self.point_transformer.transform_tangent_vec(
            cotangent_vec_a, base_point
        )
        cotangent_vec_b_ = self.point_transformer.transform_tangent_vec(
            cotangent_vec_b, base_point
        )

        res = self.space.metric.inner_coproduct(
            cotangent_vec_a, cotangent_vec_b, base_point
        )
        res_ = self.other_space.metric.inner_coproduct(
            cotangent_vec_a_, cotangent_vec_b_, base_point_
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_inner_coproduct_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        cotangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        cotangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        self.test_inner_coproduct(cotangent_vec_a, cotangent_vec_b, base_point, atol)

    def test_squared_norm(self, vector, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        vector_ = self.point_transformer.transform_tangent_vec(vector, base_point)

        res = self.space.metric.squared_norm(vector, base_point)
        res_ = self.other_space.metric.squared_norm(vector_, base_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_squared_norm_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        vector = self.data_generator.random_tangent_vec(base_point)

        self.test_squared_norm(vector, base_point, atol)

    def test_norm(self, vector, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        vector_ = self.point_transformer.transform_tangent_vec(vector, base_point)

        res = self.space.metric.norm(vector, base_point)
        res_ = self.other_space.metric.norm(vector_, base_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_norm_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        vector = self.data_generator.random_tangent_vec(base_point)

        self.test_norm(vector, base_point, atol)

    def test_normalize(self, vector, base_point, atol):
        res = self.space.metric.normalize(vector, base_point)
        res_ = self.other_space.metric.normalize(vector, base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_normalize_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        vector = self.data_generator.random_tangent_vec(base_point)

        self.test_normalize(vector, base_point, atol)

    def test_squared_dist(self, point_a, point_b, atol):
        point_a_ = self.point_transformer.transform_point(point_a)
        point_b_ = self.point_transformer.transform_point(point_b)

        res = self.space.metric.squared_dist(point_a, point_b)
        res_ = self.other_space.metric.squared_dist(point_a_, point_b_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_squared_dist_random(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        self.test_squared_dist(point_a, point_b, atol)

    def test_dist(self, point_a, point_b, atol):
        point_a_ = self.point_transformer.transform_point(point_a)
        point_b_ = self.point_transformer.transform_point(point_b)

        res = self.space.metric.dist(point_a, point_b)
        res_ = self.other_space.metric.dist(point_a_, point_b_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_dist_random(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        self.test_dist(point_a, point_b, atol)

    def test_covariant_riemann_tensor(self, base_point, atol):
        res = self.space.metric.covariant_riemann_tensor(base_point)
        res_ = self.other_space.metric.covariant_riemann_tensor(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_covariant_riemann_tensor_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_covariant_riemann_tensor(base_point, atol)

    def test_sectional_curvature(self, tangent_vec_a, tangent_vec_b, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        tangent_vec_a_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_a, base_point
        )
        tangent_vec_b_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_b, base_point
        )

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        res_ = self.other_space.metric.sectional_curvature(
            tangent_vec_a_, tangent_vec_b_, base_point_
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        self.test_sectional_curvature(tangent_vec_a, tangent_vec_b, base_point, atol)

    def test_scalar_curvature(self, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)

        res = self.space.metric.scalar_curvature(base_point)
        res_ = self.other_space.metric.scalar_curvature(base_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_scalar_curvature_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_scalar_curvature(base_point, atol)



import pytest

import geomstats.backend as gs
from geomstats.geometry.grassmannian import Grassmannian
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class StiefelTestCase(LevelSetTestCase):
    def test_to_grassmannian(self, point, expected, atol):
        projected = self.space.to_grassmannian(point)
        self.assertAllClose(projected, expected, atol=atol)

    @pytest.mark.random
    def test_to_grassmannian_belongs_to_grassmannian(self, n_points, atol):
        point = self.space.random_point(n_points)
        grass_point = self.space.to_grassmannian(point)

        grassmannian = Grassmannian(self.space.n, self.space.p, equip=False)
        res = grassmannian.belongs(grass_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)

        self.assertAllEqual(res, expected)


class StiefelStaticMethodsTestCase(TestCase):
    def test_to_grassmannian(self, point, expected, atol):
        projected = self.Space.to_grassmannian(point)
        self.assertAllClose(projected, expected, atol=atol)


class StiefelCanonicalMetricTestCase(RiemannianMetricTestCase):
    def test_retraction(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.retraction(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_lifting(self, point, base_point, expected, atol):
        res = self.space.metric.lifting(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_lifting_vec(self, n_reps, atol):
        while True:
            point = self.data_generator.random_point()
            base_point = self.data_generator.random_point()
            try:
                expected = self.space.metric.lifting(point, base_point)
                break
            except ValueError:
                # algorithm does not work for several cases
                pass

        vec_data = generate_vectorization_data(
            data=[
                dict(point=point, base_point=base_point, expected=expected, atol=atol)
            ],
            arg_names=["point", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_lifting_is_tangent(self, n_points, atol):
        while True:
            point = self.data_generator.random_point(n_points)
            base_point = self.data_generator.random_point(n_points)
            try:
                tangent_vec = self.space.metric.lifting(point, base_point)
                break
            except ValueError:
                # algorithm does not work for several cases
                pass

        res = self.space.is_tangent(tangent_vec, base_point, atol=atol)
        self.assertAllEqual(res, gs.ones(n_points, dtype=bool))

    @pytest.mark.random
    def test_lifting_after_retraction(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        point = self.space.metric.retraction(tangent_vec, base_point)
        tangent_vec_ = self.space.metric.lifting(point, base_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    @pytest.mark.random
    def test_retraction_after_lifting(self, n_points, atol):
        while True:
            point = self.data_generator.random_point(n_points)
            base_point = self.data_generator.random_point(n_points)
            try:
                tangent_vec = self.space.metric.lifting(point, base_point)
                break
            except ValueError:
                # algorithm does not work for several cases
                pass

        point_ = self.space.metric.retraction(tangent_vec, base_point)
        self.assertAllClose(point_, point, atol=atol)

    @pytest.mark.random
    def test_two_sheets_error(self, n_points):
        if self.space.n != self.space.p:
            raise ValueError("Test only defined for n=p.")

        point = self.space.random_point(n_points)
        base_point = self.space.random_point(n_points)

        det_point = gs.linalg.det(point)
        det_base = gs.linalg.det(base_point)
        if n_points > 1:
            for index in range(n_points):
                if (det_base[index] * det_point[index]) > 0.0:
                    point[index][0, :] *= -1
        else:
            if (det_base * det_point) > 0.0:
                point[0, :] *= -1.0

        with pytest.raises(ValueError):
            self.space.metric.log(point, base_point)



"""Graph Space.

Lead author: Anna Calissano.

References
----------
.. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
    “Graph Space: Geodesic Principal Components for a Population of
    Network-valued Data.” Mox report 14, 2020.
    https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
"""

import itertools
from abc import ABC, abstractmethod

import geomstats.backend as gs
from geomstats.errors import check_parameter_accepted_values
from geomstats.geometry.fiber_bundle import AlignerAlgorithm
from geomstats.geometry.group_action import PermutationAction
from geomstats.geometry.manifold import register_quotient
from geomstats.geometry.matrices import Matrices, MatricesMetric
from geomstats.geometry.stratified.quotient import Aligner, QuotientMetric
from geomstats.numerics.optimization import ScipyMinimize
from geomstats.vectorization import check_is_batch, get_batch_shape


class GraphSpaceAlignerAlgorithm(AlignerAlgorithm, ABC):
    """Base class for graph space numerical aligner.

    Attributes
    ----------
    total_space : GraphSpace
        Set with quotient structure.
    perm_ : array-like, shape=[..., n_nodes]
        Node permutations where in position i we have the value j meaning
        the node i should be permuted with node j.
    """

    def __init__(self, total_space):
        super().__init__(total_space)
        self.perm_ = None

    def _get_opt_perm(self, point, base_point):
        """Get optimal element of the group.

        Parameters
        ----------
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.
        base_point : array-like, shape=[..., n_nodes, n_nodes]
            Base graph.

        Returns
        -------
        perm : array-like, shape=[..., n_nodes]
            Optimal permutation group element.
        """
        is_batch = check_is_batch(self._total_space.point_ndim, point, base_point)
        if is_batch:
            if point.ndim != base_point.ndim:
                point, base_point = gs.broadcast_arrays(point, base_point)
            return gs.stack(
                [
                    self._get_opt_perm_single(point_, base_point_)
                    for point_, base_point_ in zip(point, base_point)
                ]
            )
        return self._get_opt_perm_single(point, base_point)

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.
        base_point : array-like, shape=[..., n_nodes, n_nodes]
            Reference graph.

        Returns
        -------
        aligned_point : array-like, shape=[..., n_nodes, n_nodes]
            Aligned graph.
        """
        self.perm_ = self._get_opt_perm(point, base_point)
        return self._total_space.group_action(self.perm_, point)


class FAQAligner(GraphSpaceAlignerAlgorithm):
    """Fast Quadratic Assignment for graph matching (or network alignment).

    References
    ----------
    .. [Vogelstein2015] Vogelstein JT, Conroy JM, Lyzinski V, Podrazik LJ,
        Kratzer SG, Harley ET, Fishkind DE, Vogelstein RJ, Priebe CE.
        “Fast approximate quadratic programming for graph matching.“
        PLoS One. 2015 Apr 17; doi: 10.1371/journal.pone.0121002.
    """

    def _get_opt_perm_single(self, point, base_point):
        """Get optimal element of the group.

        Parameters
        ----------
        point : array-like, shape=[n_nodes, n_nodes]
            Graph to align.
        base_point : array-like, shape=[n_nodes, n_nodes]
            Base graph.

        Returns
        -------
        perm : array-like, shape=[n_nodes]
            Optimal permutation group element.
        """
        return gs.array(
            gs.linalg.quadratic_assignment(
                base_point, point, options={"maximize": True}
            )
        )


class ExhaustiveAligner(GraphSpaceAlignerAlgorithm):
    """Brute force exact alignment.

    Exact Alignment obtained by exploring the whole permutation group.

    Parameters
    ----------
    total_space : GraphSpace
        Set with quotient structure.

    Notes
    -----
    Not recommended for large `n_nodes`.
    """

    def __init__(self, total_space):
        super().__init__(total_space)
        n_nodes = total_space.n_nodes
        self._perms = gs.array(list(itertools.permutations(range(n_nodes), n_nodes)))

    def _get_opt_perm_single(self, point, base_point):
        """Get optimal element of the group.

        Parameters
        ----------
        point : array-like, shape=[n_nodes, n_nodes]
            Graph to align.
        base_point : array-like, shape=[n_nodes, n_nodes]
            Base graph.

        Returns
        -------
        perm : array-like, shape=[n_nodes]
            Optimal permutation group element.
        """
        orbit = self._total_space.group_action(self._perms, point)
        sdists = self._total_space.metric.squared_dist(base_point, orbit)
        return self._perms[gs.argmin(sdists)]


class PointToGeodesicAlignerBase(ABC):
    """Base class for point to geodesic aligner.

    Parameters
    ----------
    total_space : GraphSpace
        Set with quotient structure.

    Attributes
    ----------
    perm_ : array-like, shape=[..., n_nodes]
        Node permutations where in position i we have the value j meaning
        the node i should be permuted with node j.
    """

    def __init__(self, total_space):
        self._total_space = total_space
        self.perm_ = None

    @abstractmethod
    def align(self, geodesic, point):
        """Class for the alignment of the geodesic with respect to a point."""

    @abstractmethod
    def dist(self, geodesic, point):
        """Class to compute distance between the geodesic with respect to a point."""


class PointToGeodesicAligner(PointToGeodesicAlignerBase):
    """Class for the alignment of the points with respect to a geodesic.

    Implementing the algorithm in [Huckemann2010]_ to select an optimal alignment to a
    point with respect to a geodesic. The algorithm sample discrete set of n_points
    along the geodesic between [s_min, s_max] and find the permutation that gets closer
    to the datapoints along the geodesic.

    Parameters
    ----------
    total_space : GraphSpace
        Set with quotient structure.
    s_min : float
        Minimum value of the domain to sample along the geodesics.
    s_max : float
        Minimum value of the domain to sample along the geodesics.
    n_grid: int
        Number of points to sample between s_min and s_max.

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    .. [Huckemann2010] Huckemann, S., Hotz, T., Munk, A.
        "Intrinsic shape analysis: Geodesic PCA for Riemannian manifolds modulo
        isometric Lie group actions." Statistica Sinica, 1-58, 2010.
    """

    def __init__(self, total_space, s_min=0.0, s_max=1.0, n_grid=10):
        super().__init__(total_space)
        self.s_min = s_min
        self.s_max = s_max
        self.n_grid = n_grid
        self._s_grid = None

    def __setattr__(self, attr_name, value):
        """Set attributes."""
        if attr_name in ["s_min", "s_max", "n_points"]:
            self._s_grid = None

        return object.__setattr__(self, attr_name, value)

    def _discretize_s(self):
        """Compute the domain distretization."""
        return gs.linspace(self.s_min, self.s_max, num=self.n_grid)

    @property
    def s_grid(self):
        """Save the domain discretization."""
        if self._s_grid is None:
            self._s_grid = self._discretize_s()

        return self._s_grid

    def _compute_dists(self, geodesic, point):
        """Compute the distance between the geodesic and the point.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        dists : array-like, shape=[..., n_grid]
        min_dists_idx : array-like, shape=[...,]
        aligned_points : array-like, shape=[..., n_grid, n_nodes, n_nodes]
        """
        total_space = self._total_space
        geodesic_points = geodesic(self.s_grid)
        geo_shape = (self.n_grid,) + total_space.shape

        geo_batch_shape = get_batch_shape(3, geodesic_points)
        point_batch_shape = get_batch_shape(2, point)
        if point_batch_shape or geo_batch_shape:
            batch_shape = point_batch_shape or geo_batch_shape
        else:
            batch_shape = ()

        point = gs.broadcast_to(point, (self.n_grid,) + batch_shape + total_space.shape)
        if point_batch_shape:
            point = gs.moveaxis(point, 0, -3)

        if batch_shape and not geo_batch_shape:
            geodesic_points = gs.broadcast_to(geodesic_points, batch_shape + geo_shape)

        flat_point = gs.reshape(point, (-1,) + total_space.shape)
        flat_geodesic_s = gs.reshape(geodesic_points, (-1,) + total_space.shape)

        aligned_flat_points = total_space.aligner.align(flat_point, flat_geodesic_s)
        flat_dists = total_space.metric.dist(flat_geodesic_s, aligned_flat_points)

        perm_ = total_space.aligner.perm_
        total_space.aligner.align_algo.perm_ = gs.reshape(
            perm_, batch_shape + (self.n_grid, total_space.n_nodes)
        )

        dists = gs.reshape(
            flat_dists,
            batch_shape + (self.n_grid,),
        )

        min_dists_idx = gs.argmin(dists, axis=-1)

        aligned_points = gs.reshape(
            aligned_flat_points,
            batch_shape + geo_shape,
        )

        return dists, min_dists_idx, aligned_points

    def dist(self, geodesic, point):
        """Compute the distance between the geodesic and the point.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        dist : array-like, shape=[..., n_nodes]
            Distance between the point and the geodesic.

        Notes
        -----
        Due to the discrete nature of the method, distance is not very accurate.
        """
        dists, min_dists_idx, _ = self._compute_dists(geodesic, point)
        slc = []
        for n in dists.shape[:-1]:
            slc.append(gs.arange(n))
        slc.append(min_dists_idx)
        return dists[tuple(slc)]

    def align(self, geodesic, point):
        """Align the graph to the geodesic.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        aligned_point : array-like, shape=[..., n_nodes, n_nodes]
            Permuted graph as to be aligned with respect to the geodesic.
        """
        _, min_dists_idx, aligned_points = self._compute_dists(geodesic, point)
        slc = []
        for n in aligned_points.shape[:-3]:
            slc.append(gs.arange(n))
        slc.append(min_dists_idx)
        self.perm_ = self._total_space.aligner.perm_[tuple(slc)]

        slc.extend([slice(None), slice(None)])
        return aligned_points[tuple(slc)]


class _GeodesicToPointAligner(PointToGeodesicAlignerBase):
    """Class for the alignment of the points with respect to a geodesic.

    Solves a 1d optimization problem.

    Parameters
    ----------
    total_space : GraphSpace
        Set with quotient structure.
    save_opt_res : bool
        Whether to save optimization results.
    """

    def __init__(self, total_space, save_opt_res=False):
        super().__init__(total_space)

        self.save_opt_res = save_opt_res
        self.minimizer = ScipyMinimize(method="BFGS")

        self.opt_results_ = None

    def _objective_single(self, param, geodesic, point):
        """Objective function.

        Parameters
        ----------
        param : array-like, shape=[1,]
            Parameter along the geodesic.
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        dist : array-like, shape=[]
            Dist from point to geodesic.
        """
        geodesic_point = geodesic(param)
        if geodesic_point.ndim > 3:
            raise NotImplementedError("Cannot handle more than one geodesic at time")

        geodesic_point = gs.squeeze(geodesic_point, axis=0)
        return self._total_space.quotient.metric.squared_dist(geodesic_point, point)

    def _optimize_single(self, geodesic, point):
        """Solution of optimization problem.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        res : OptimizationResult
            Result of optimization.
        """

        def objective(param):
            return self._objective_single(param, geodesic=geodesic, point=point)

        return self.minimizer.minimize(
            objective,
            x0=0.0,
        )

    def squared_dist(self, geodesic, point, return_perm=False):
        """Compute the distance between the geodesic and the point.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.
        return_perm : bool
            If to return optimal permutations.

        Returns
        -------
        sdist : array-like, shape=[...]
            Squared distance between point and geodesic.
        perm : array-like, shape=[..., n_nodes]
            Optimal permutations.
        """
        batch_shape = get_batch_shape(self._total_space.point_ndim, point)

        if not batch_shape:
            point = gs.expand_dims(point, axis=0)

        perms = []
        min_sdists = []
        opt_results = []
        for point_ in point:
            res = self._optimize_single(geodesic, point_)
            perms.append(self._total_space.aligner.perm_)
            min_sdists.append(res.fun)

            opt_results.append(res)

        if not batch_shape:
            min_sdists = min_sdists[0]
            perms = perms[0]
            opt_results = opt_results[0]

        if self.save_opt_res:
            self.opt_results_ = opt_results

        min_sdists = gs.array(min_sdists)
        if return_perm:
            return min_sdists, gs.array(perms)

        return min_sdists

    def dist(self, geodesic, point):
        """Compute the distance between the geodesic and the point.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        dist : array-like, shape=[..., n_nodes]
            Distance between the point and the geodesic.
        """
        sdist, _ = self.squared_dist(geodesic, point, return_perm=True)
        return gs.sqrt(sdist)

    def align(self, geodesic, point):
        """Align the graph to the geodesic.

        Parameters
        ----------
        geodesic : function
            Geodesic function in GraphSpace.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        aligned_point : array-like, shape=[..., n_nodes, n_nodes]
            Permuted graph as to be aligned with respect to the geodesic.
        """
        _, self.perm_ = self.squared_dist(geodesic, point, return_perm=True)
        return self._total_space.group_action(self.perm_, point)


class GraphSpace(Matrices):
    r"""Class for the Graph Space.

    Graph Space to analyse populations of labelled and unlabelled graphs.
    The space focuses on graphs with scalar euclidean attributes on nodes and edges,
    with a finite number of nodes and both directed and undirected edges.
    For undirected graphs, use symmetric adjacency matrices. The space is a quotient
    space obtained by applying the permutation action of nodes to the space
    of adjacency matrices. Notice that for computation reasons the module works with
    both the `gs.array` representation of graph and the `GraphPoint` representation.

    Points are represented by :math:`nodes \times nodes` adjacency matrices.
    Both the array input and the Graph Point type input work.

    Parameters
    ----------
    n_nodes : int
        Number of graph nodes
    total_space : space
        Total Space before applying the permutation action. Default: Euclidean Space.

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    .. [Jain2009] Jain, B., Obermayer, K.
        "Structure Spaces." Journal of Machine Learning Research, 10(11), 2009.
        https://www.jmlr.org/papers/volume10/jain09a/jain09a.pdf
    """

    def __init__(self, n_nodes, equip=True):
        self.n_nodes = n_nodes
        super().__init__(n_nodes, n_nodes, equip=equip)

    def new(self, equip=True):
        """Create manifold with same parameters."""
        return GraphSpace(n_nodes=self.n_nodes, equip=equip)

    def equip_with_group_action(self, group_action="permutations"):
        """Equip manifold with group action."""
        if group_action == "permutations":
            group_action = PermutationAction()

        return super().equip_with_group_action(group_action)


class GraphSpaceAligner(Aligner):
    """Graph space aligner.

    Parameters
    ----------
    total_space : GraphSpace
        Set with quotient structure.
    align_algo : GraphSpaceAlignerAlgorithm
        Algorihtm performing alignment.
    """

    MAP_ALIGNER = {
        "FAQ": FAQAligner,
        "exhaustive": ExhaustiveAligner,
    }

    def __init__(self, total_space, align_algo=None):
        super().__init__(total_space=total_space, align_algo=align_algo)
        if align_algo is None:
            align_algo = self.set_alignment_algorithm()
        self.point_to_geodesic_aligner = self.set_point_to_geodesic_aligner()

    @property
    def perm_(self):
        """Optimal node permutations.

        Returns
        -------
        perm_ : array-like, shape=[..., n_nodes]
            Node permutations where in position i we have the value j meaning
        the node i should be permuted with node j.
        """
        return self.align_algo.perm_

    def set_alignment_algorithm(self, align_algo="FAQ", **kwargs):
        """Set the aligning strategy.

        GraphSpace metric relies on alignment. In this module we propose the
        the FAQ graph matching by [Vogelstein2015]_, and
        exhaustive aligner which explores the whole permutation group.

        Parameters
        ----------
        align_algo : str or GraphSpaceAlignerAlgorithm
            'FAQ': Fast Quadratic Assignment - only compatible with Frobenius norm,
            'exhaustive': all group exhaustive search
        """
        if isinstance(align_algo, str):
            check_parameter_accepted_values(
                align_algo, "align_algo", list(self.MAP_ALIGNER.keys())
            )

            aligner_algorithm = self.MAP_ALIGNER.get(align_algo)(
                self._total_space, **kwargs
            )

        self.align_algo = aligner_algorithm
        return self.align_algo

    def set_point_to_geodesic_aligner(self, aligner="default", **kwargs):
        """Set the alignment between a point and a geodesic.

        Following the geodesic to point alignment in [Calissano2020]_ and
        [Huckemann2010]_, this function defines the parameters [s_min, s_max] and
        the number of points to sample in the domain.

        Parameters
        ----------
        aligner: BasePointToGeodesicAligner
        s_min : float
            Minimum value of the domain to sample along the geodesics.
        s_max : float
            Minimum value of the domain to sample along the geodesics.
        n_points: int
            Number of points to sample between s_min and s_max.
        """
        if aligner == "default":
            kwargs.setdefault("s_min", -1.0)
            kwargs.setdefault("s_max", 1.0)
            kwargs.setdefault("n_grid", 10)
            aligner = PointToGeodesicAligner(self._total_space, **kwargs)

        self.point_to_geodesic_aligner = aligner
        return self.point_to_geodesic_aligner

    def align_point_to_geodesic(self, geodesic, point):
        """Align point to a geodesic.

        Using the selected alignment technique, it returns the aligned
        point as optimally aligned to the geodesic.

        Parameters
        ----------
        geodesic : function
            Geodesic.
        point : array-like, shape=[..., n_nodes, n_nodes]
            Graph to align.

        Returns
        -------
        permuted_graph: list, shape = [..., n_nodes, n_nodes]
        """
        return self.point_to_geodesic_aligner.align(geodesic, point)


class GraphSpaceQuotientMetric(QuotientMetric):
    r"""Class for the Graph Space Metric.

    Every metric :math:`d: X \times X \rightarrow \mathbb{R}` on the total space of
    adjacency matrices can descend onto the quotient space as a pseudo-metric:
    :math:`d([x_1],[x_2]) = min_{t\in T} d_X(x_1, t^Tx_2t)`. The metric relies on the
    total space metric and an alignment procedure, i.e., Graph Matching or Networks
    alignment procedure. Metric, alignment, geodesics, and alignment with respect to
    a geodesic are defined. By default, the alignment is FAQ and the total
    space metric is the Frobenious norm.

    Parameters
    ----------
    space : GraphSpace
        GraphSpace object.

    References
    ----------
    .. [Calissano2020]  Calissano, A., Feragen, A., Vantini, S.
        “Graph Space: Geodesic Principal Components for a Population of
        Network-valued Data.” Mox report 14, 2020.
        https://mox.polimi.it/reports-and-theses/publication-results/?id=855.
    .. [Jain2009]  Jain, B., Obermayer, K.
        "Structure Spaces." Journal of Machine Learning Research 10.11 (2009).
        https://www.jmlr.org/papers/v10/jain09a.html.
    """


register_quotient(
    Space=GraphSpace,
    Metric=MatricesMetric,
    GroupAction=PermutationAction,
    FiberBundle=GraphSpaceAligner,
    QuotientMetric=GraphSpaceQuotientMetric,
)



"""Statistical Manifold of Dirichlet distributions with the Fisher metric.

Lead author: Alice Le Brigant.
"""

import math

import numpy as np
from scipy.optimize import minimize
from scipy.stats import dirichlet

import geomstats.backend as gs
from geomstats.algebra_utils import from_vector_to_diagonal_matrix
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyMultivariateRandomVariable,
)
from geomstats.numerics.bvp import ScipySolveBVP
from geomstats.numerics.geodesic import ExpODESolver, LogODESolver
from geomstats.numerics.ivp import ScipySolveIVP
from geomstats.vectorization import repeat_out


class DirichletDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    """Class for the manifold of Dirichlet distributions.

    This is Dirichlet = :math:`(R_+^*)^dim`, the positive quadrant of the
    dim-dimensional Euclidean space.

    Attributes
    ----------
    dim : int
        Dimension of the manifold of Dirichlet distributions.
    """

    def __init__(self, dim, equip=True):
        super().__init__(
            dim=dim,
            support_shape=(dim,),
            embedding_space=Euclidean(dim=dim, equip=False),
            equip=equip,
        )
        self._scp_rv = DirichletRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return DirichletMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold of Dirichlet distributions.

        Check that point defines parameters for a Dirichlet distributions,
        i.e. belongs to the positive quadrant of the Euclidean space.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to be checked.
        atol : float
            Tolerance to evaluate positivity.
            Optional, default: gs.atol

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean indicating whether point represents a Dirichlet
            distribution.
        """
        belongs = point.shape[-1] == self.dim
        if not belongs:
            return gs.zeros(point.shape[:-1], dtype=bool)

        return gs.all(point >= -atol, axis=-1)

    def random_point(self, n_samples=1, bound=5.0):
        """Sample parameters of Dirichlet distributions.

        The uniform distribution on [0, bound]^dim is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of the square where the Dirichlet parameters are sampled.
            Optional, default: 5.

        Returns
        -------
        samples : array-like, shape=[..., dim]
            Sample of points representing Dirichlet distributions.
        """
        size = (self.dim,) if n_samples == 1 else (n_samples, self.dim)
        return bound * gs.random.rand(*size)

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        The last coordinate is floored to `gs.atol` if it is negative.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in ambient space.
        atol : float
            Tolerance to evaluate positivity.

        Returns
        -------
        projected : array-like, shape=[..., dim]
            Projected point.
        """
        return gs.where(point < atol, atol, point)

    def sample(self, point, n_samples=1):
        """Sample from the Dirichlet distribution.

        Sample from the Dirichlet distribution with parameters provided
        by point. This gives n_samples points in the simplex.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a Dirichlet distribution.
        n_samples : int
            Number of points to sample for each set of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, dim]
            Sample from the Dirichlet distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the Dirichlet
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a beta distribution.

        Returns
        -------
        pdf : function
            Probability density function of the Dirichlet distribution with
            parameters provided by point.
        """
        return lambda x: self._scp_rv.pdf(x, point=point)


class DirichletMetric(RiemannianMetric):
    """Class for the Fisher information metric on Dirichlet distributions."""

    def __init__(self, space):
        super().__init__(space=space)

        self.log_solver = LogODESolver(
            space, n_nodes=1000, integrator=ScipySolveBVP(max_nodes=1000)
        )
        self.exp_solver = ExpODESolver(space, integrator=ScipySolveIVP(method="LSODA"))

    def metric_matrix(self, base_point):
        """Compute the inner-product matrix.

        Compute the inner-product matrix of the Fisher information metric
        at the tangent space at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        base_point = gs.to_ndarray(base_point, to_ndim=2)
        n_points = base_point.shape[0]

        mat_ones = gs.ones((n_points, self._space.dim, self._space.dim))
        poly_sum = gs.polygamma(1, gs.sum(base_point, -1))
        mat_diag = from_vector_to_diagonal_matrix(gs.polygamma(1, base_point))

        mat = mat_diag - gs.einsum("i,ijk->ijk", poly_sum, mat_ones)
        return gs.squeeze(mat)

    def christoffels(self, base_point):
        """Compute the Christoffel symbols.

        Compute the Christoffel symbols of the Fisher information metric.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        christoffels : array-like, shape=[..., dim, dim, dim]
            Christoffel symbols, with the contravariant index on
            the first dimension.
            :math:`christoffels[..., i, j, k] = Gamma^i_{jk}`

        References
        ----------
        .. [LPP2021] A. Le Brigant, S. C. Preston, S. Puechmorel. Fisher-Rao
            geometry of Dirichlet Distributions. Differential Geometry
            and its Applications, 74, 101702, 2021.
        """
        base_point = gs.to_ndarray(base_point, to_ndim=2)
        n_points = base_point.shape[0]

        def coefficients(ind_k):
            """Christoffel symbols for contravariant index ind_k."""
            param_k = base_point[..., ind_k]
            param_sum = gs.sum(base_point, -1)
            c1 = (
                1
                / gs.polygamma(1, param_k)
                / (
                    1 / gs.polygamma(1, param_sum)
                    - gs.sum(1 / gs.polygamma(1, base_point), -1)
                )
            )
            c2 = -c1 * gs.polygamma(2, param_sum) / gs.polygamma(1, param_sum)

            mat_ones = gs.ones((n_points, self._space.dim, self._space.dim))
            mat_diag = from_vector_to_diagonal_matrix(
                -gs.polygamma(2, base_point) / gs.polygamma(1, base_point)
            )
            arrays = [
                gs.zeros((1, ind_k)),
                gs.ones((1, 1)),
                gs.zeros((1, self._space.dim - ind_k - 1)),
            ]
            vec_k = gs.tile(gs.hstack(arrays), (n_points, 1))
            val_k = gs.polygamma(2, param_k) / gs.polygamma(1, param_k)
            vec_k = gs.einsum("i,ij->ij", val_k, vec_k)
            mat_k = from_vector_to_diagonal_matrix(vec_k)

            mat = (
                gs.einsum("i,ijk->ijk", c2, mat_ones)
                - gs.einsum("i,ijk->ijk", c1, mat_diag)
                + mat_k
            )

            return 1 / 2 * mat

        christoffels = []
        for ind_k in range(self._space.dim):
            christoffels.append(coefficients(ind_k))
        christoffels = gs.stack(christoffels, 1)

        return gs.squeeze(christoffels)

    def jacobian_christoffels(self, base_point):
        """Compute the Jacobian of the Christoffel symbols.

        Compute the Jacobian of the Christoffel symbols of the
        Fisher information metric.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        jac : array-like, shape=[..., dim, dim, dim, dim]
            Jacobian of the Christoffel symbols.
            :math:`jac[..., i, j, k, l] = dGamma^i_{jk} / dx_l`
        """
        dim = self._space.dim
        n_dim = base_point.ndim
        param = gs.transpose(base_point)
        sum_param = gs.sum(param, 0)
        term_1 = 1 / gs.polygamma(1, param)
        term_2 = 1 / gs.polygamma(1, sum_param)
        term_3 = -gs.polygamma(2, param) / gs.polygamma(1, param) ** 2
        term_4 = -gs.polygamma(2, sum_param) / gs.polygamma(1, sum_param) ** 2
        term_5 = term_3 / term_1
        term_6 = term_4 / term_2
        term_7 = (
            gs.polygamma(2, param) ** 2
            - gs.polygamma(1, param) * gs.polygamma(3, param)
        ) / gs.polygamma(1, param) ** 2
        term_8 = (
            gs.polygamma(2, sum_param) ** 2
            - gs.polygamma(1, sum_param) * gs.polygamma(3, sum_param)
        ) / gs.polygamma(1, sum_param) ** 2
        term_9 = term_2 - gs.sum(term_1, 0)

        jac_1 = term_1 * term_8 / term_9
        jac_1_mat = gs.squeeze(gs.tile(jac_1, (dim, dim, dim, 1, 1)))
        jac_2 = (
            -term_6 / term_9**2 * gs.einsum("j...,i...->ji...", term_4 - term_3, term_1)
        )
        jac_2_mat = gs.squeeze(gs.tile(jac_2, (dim, dim, 1, 1, 1)))
        jac_3 = term_3 * term_6 / term_9
        jac_3_mat = gs.transpose(from_vector_to_diagonal_matrix(gs.transpose(jac_3)))
        jac_3_mat = gs.squeeze(gs.tile(jac_3_mat, (dim, dim, 1, 1, 1)))
        jac_4 = (
            1
            / term_9**2
            * gs.einsum("k...,j...,i...->kji...", term_5, term_4 - term_3, term_1)
        )
        jac_4_mat = gs.transpose(from_vector_to_diagonal_matrix(gs.transpose(jac_4)))
        jac_5 = -gs.einsum("j...,i...->ji...", term_7, term_1) / term_9
        jac_5_mat = from_vector_to_diagonal_matrix(gs.transpose(jac_5))
        jac_5_mat = gs.transpose(from_vector_to_diagonal_matrix(jac_5_mat))
        jac_6 = -gs.einsum("k...,j...->kj...", term_5, term_3) / term_9
        jac_6_mat = gs.transpose(from_vector_to_diagonal_matrix(gs.transpose(jac_6)))
        jac_6_mat = (
            gs.transpose(
                from_vector_to_diagonal_matrix(gs.transpose(jac_6_mat, [0, 1, 3, 2])),
                [0, 1, 3, 4, 2],
            )
            if n_dim > 1
            else from_vector_to_diagonal_matrix(jac_6_mat)
        )
        jac_7 = -from_vector_to_diagonal_matrix(gs.transpose(term_7))
        jac_7_mat = from_vector_to_diagonal_matrix(jac_7)
        jac_7_mat = gs.transpose(from_vector_to_diagonal_matrix(jac_7_mat))

        jac = (
            1
            / 2
            * (
                jac_1_mat
                + jac_2_mat
                + jac_3_mat
                + jac_4_mat
                + jac_5_mat
                + jac_6_mat
                + jac_7_mat
            )
        )

        return (
            gs.transpose(jac, [3, 1, 0, 2])
            if n_dim == 1
            else gs.transpose(jac, [4, 3, 1, 0, 2])
        )

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base point onto
        its image.
        In the case of the hyperbolic space, it does not depend on the base point and
        is infinite everywhere, because of the negative curvature.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)

    def _approx_geodesic_bvp(
        self,
        initial_point,
        end_point,
        degree=5,
        method="BFGS",
        n_times=200,
        jac_on=True,
    ):
        """Solve approximation of the geodesic boundary value problem.

        The space of solutions is restricted to curves whose coordinates are
        polynomial functions of time. The boundary value problem is solved by
        minimizing the energy among all such curves starting from initial_point
        and ending at end_point, i.e. curves t -> (x_1(t),...,x_n(t)) where x_i
        are polynomial functions of time t, such that (x_1(0),..., x_n(0)) is
        initial_point and (x_1(1),..., x_n(1)) is end_point. The parameterized
        curve is computed at n_times discrete times.

        Parameters
        ----------
        initial_point : array-like, shape=(dim,)
            Starting point of the geodesic.
        end_point : array-like, shape=(dim,)
            End point of the geodesic.
        degree : int
            Degree of the coordinates' polynomial functions of time.
        method : str
            Minimization method to use in scipy.optimize.minimize.
        n_times : int
            Number of sample times.
        jac_on : bool
            If jac_on=True, use the Jacobian of the energy cost function in
            scipy.optimize.minimize.

        Returns
        -------
        dist : float
            Length of the polynomial approximation of the geodesic.
        curve : array-like, shape=(n_times, dim)
            Polynomial approximation of the geodesic.
        velocity : array-like, shape=(n_times, dim)
            Velocity of the polynomial approximation of the geodesic.
        """

        def cost_fun(param):
            """Compute the energy of the polynomial curve defined by param.

            Parameters
            ----------
            param : array-like, shape=(degree - 1, dim)
                Parameters of the curve coordinates' polynomial functions of time.

            Returns
            -------
            energy : float
                Energy of the polynomial approximation of the geodesic.
            length : float
                Length of the polynomial approximation of the geodesic.
            curve : array-like, shape=(n_times, dim)
                Polynomial approximation of the geodesic.
            velocity : array-like, shape=(n_times, dim)
                Velocity of the polynomial approximation of the geodesic.
            """
            last_coef = end_point - initial_point - gs.sum(param, axis=0)
            coef = gs.vstack((initial_point, param, last_coef))

            t = gs.linspace(0.0, 1.0, n_times)
            t_curve = [t**i for i in range(degree + 1)]
            t_curve = gs.stack(t_curve)
            curve = gs.einsum("ij,ik->kj", coef, t_curve)

            t_velocity = [i * t ** (i - 1) for i in range(1, degree + 1)]
            t_velocity = gs.stack(t_velocity)
            velocity = gs.einsum("ij,ik->kj", coef[1:], t_velocity)

            if curve.min() < 0:
                return np.inf, np.inf, curve, np.nan

            velocity_sqnorm = self.squared_norm(vector=velocity, base_point=curve)
            length = gs.sum(velocity_sqnorm ** (1 / 2)) / n_times
            energy = gs.sum(velocity_sqnorm) / n_times
            return energy, length, curve, velocity

        def cost_jacobian(param):
            """Compute the jacobian of the cost function at polynomial curve.

            Parameters
            ----------
            param : array-like, shape=(degree - 1, dim)
                Parameters of the curve coordinates' polynomial functions of time.

            Returns
            -------
            jac : array-like, shape=(dim * (degree - 1),)
                Jacobian of the cost function at polynomial curve.
            """
            last_coef = end_point - initial_point - gs.sum(param, 0)
            coef = gs.vstack((initial_point, param, last_coef))

            t = gs.linspace(0.0, 1.0, n_times)
            t_position = [t**i for i in range(degree + 1)]
            t_position = gs.stack(t_position)
            position = gs.einsum("ij,ik->kj", coef, t_position)

            t_velocity = [i * t ** (i - 1) for i in range(1, degree + 1)]
            t_velocity = gs.stack(t_velocity)
            velocity = gs.einsum("ij,ik->kj", coef[1:], t_velocity)

            fac1 = gs.stack(
                [
                    k * t ** (k - 1) - degree * t ** (degree - 1)
                    for k in range(1, degree)
                ]
            )
            fac2 = gs.stack([t**k - t**degree for k in range(1, degree)])
            fac3 = (velocity * gs.polygamma(1, position)).T - gs.sum(
                velocity, 1
            ) * gs.polygamma(1, gs.sum(position, 1))
            fac4 = (velocity**2 * gs.polygamma(2, position)).T - gs.sum(
                velocity, 1
            ) ** 2 * gs.polygamma(2, gs.sum(position, 1))

            cost_jac = (
                2 * gs.einsum("ij,kj->ik", fac1, fac3)
                + gs.einsum("ij,kj->ik", fac2, fac4)
            ) / n_times
            return cost_jac.T.reshape(dim * (degree - 1))

        def f2minimize(x):
            """Compute function to minimize."""
            param = gs.transpose(x.reshape((dim, degree - 1)))
            res = cost_fun(param)
            return res[0]

        def jacobian(x):
            """Compute jacobian of the function to minimize."""
            param = gs.transpose(x.reshape((dim, degree - 1)))
            return cost_jacobian(param)

        dim = initial_point.shape[0]
        x0 = gs.ones(dim * (degree - 1))
        jac = jacobian if jac_on else None
        sol = minimize(f2minimize, x0, method=method, jac=jac)
        opt_param = sol.x.reshape((dim, degree - 1)).T
        _, dist, curve, velocity = cost_fun(opt_param)

        return dist, curve, velocity


class DirichletRandomVariable(ScipyMultivariateRandomVariable):
    """A Dirichlet random variable."""

    def __init__(self, space):
        pdf = lambda x, point: dirichlet.pdf(gs.transpose(x), point)
        super().__init__(space, dirichlet.rvs, pdf)
